"""
Recreates line 4 of Table 1 from
Hudson, R. R., and N. L. Kaplan. 1995.
“Deleterious Background Selection with Recombination.”
Genetics 141 (4): 1605–17.

The physical layout of the genome follows their Figure 1.
The fact that the neutral region is non-recombining
is a major simplification for the analysis.  By definition,
a single tree describes the history of that region.  If that
were not the case, we would have to take the weighted
mean of summaries of trees, with tree lengths (relative
to total genome length) as the weights.
"""
import fwdpy11
import numpy as np
import sys
import concurrent.futures

GENOME_LENGTH = 1.0
R = 0.04
U = 0.08
N = 1600
# number of diploids to sample
NSAM = 10


def runsim(argtuple):
    seed = argtuple

    rng = fwdpy11.GSLrng(seed)

    pdict = {'gvalue': fwdpy11.Multiplicative(2.),
             'rates': (0., U/2., R),  # The U/2. is from their eqn. 2.
             'nregions': [],
             'sregions': [fwdpy11.ConstantS(0, 1./3., 1, -0.02, 1.),
                          fwdpy11.ConstantS(2./3., 1., 1, -0.02, 1.)],
             'recregions': [fwdpy11.Region(0, 1./3., 1),
                            fwdpy11.Region(2./3., 1., 1)],
             'demography': np.array([N]*20*N, dtype=np.uint32)
             }
    params = fwdpy11.ModelParams(**pdict)

    pop = fwdpy11.DiploidPopulation(N, GENOME_LENGTH)

    fwdpy11.evolvets(rng, pop, params, 100, suppress_table_indexing=True)

    rdips = np.random.choice(N, NSAM, replace=False)
    md = np.array(pop.diploid_metadata, copy=False)
    rdip_nodes = md['nodes'][rdips].flatten()
    nodes = np.array(pop.tables.nodes, copy=False)
    # Only visit trees spanning the
    # mutation-free segment of the genome
    tv = fwdpy11.TreeIterator(pop.tables, rdip_nodes, begin=1./3., end=2./3.)
    plist = np.zeros(len(nodes), dtype=np.int8)
    sum_pairwise_tmrca = 0
    for t in tv:
        for i in range(len(rdip_nodes)-1):
            u = rdip_nodes[i]
            while u != fwdpy11.NULL_NODE:
                plist[u] = 1
                u = t.parent(u)
            for j in range(i+1, len(rdip_nodes)):
                u = rdip_nodes[j]
                while u != fwdpy11.NULL_NODE:
                    if plist[u] == 1:
                        sum_pairwise_tmrca += 2 * \
                            (pop.generation-nodes['time'][u])
                        u = fwdpy11.NULL_NODE
                    else:
                        u = t.parent(u)
            plist.fill(0)
    return 2*sum_pairwise_tmrca/(len(rdip_nodes)*(len(rdip_nodes)-1))


if __name__ == "__main__":
    seed = int(sys.argv[1])
    nreps = int(sys.argv[2])
    np.random.seed(seed)
    seeds = []
    for i in range(nreps):
        candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        while candidate in seeds:
            candidate = np.random.randint(0, np.iinfo(np.uint32).max)
        seeds.append(candidate)

    mean_pairwise_coalescence_times = np.zeros(nreps)
    mean_pairwise_coalescence_times.fill(np.nan)
    idx = 0
    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = {executor.submit(runsim, (i)) for i in seeds}
        for future in concurrent.futures.as_completed(futures):
            scaled_mean_pairwise_coalescence_time = future.result()/(4*N)
            mean_pairwise_coalescence_times[idx] = scaled_mean_pairwise_coalescence_time
            idx += 1
    mean_pairwise_coalescence_times
    print(mean_pairwise_coalescence_times.mean(),
          mean_pairwise_coalescence_times.std()/np.sqrt(nreps))



"""
Plots the mean genetic value for all time periods
present in the tree sequence.

Usage: python3 plot_genetic_values_from_tree_sequence.py filename

filename should contain a pickled population output by SlocusPopGSSmo.py
"""
import pickle
import lzma
import sys
import numpy as np
import seaborn as sns

with lzma.open(sys.argv[1], 'rb') as f:
    pop = pickle.load(f)

nodes = np.array(pop.tables.nodes, copy=False)

times = []
mean_trait_values = []
for t, n, m in pop.sample_timepoints():
    times.append(t)
    mean_trait_values.append(m['g'].mean())

# plot the time vs trait value
xyplot = sns.lineplot(times, mean_trait_values)
xyplot.set_xlabel("Time (generations")
xyplot.set_ylabel("Mean trait value")
xyplot.get_figure().savefig("mean_trait_value_from_tree_sequences.png",
                            dpi=300)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


def mslike(pop, **kwargs):
    """
    Function to establish default parameters
    for a single-locus simulation for standard pop-gen
    modeling scenarios.

    :params pop: An instance of :class:`fwdpy11.DiploidPopulation`
    :params kwargs: Keyword arguments.
    """
    import fwdpy11
    if isinstance(pop, fwdpy11.DiploidPopulation) is False:
        raise ValueError("incorrect pop type: " + str(type(pop)))
    defaults = {'simlen': 10*pop.N,
                'beg': 0.0,
                'end': 1.0,
                'theta': 100.0,
                'pneutral': 1.0,
                'rho': 100.0,
                'dfe': None
                }
    for key, value in kwargs.items():
        if key in defaults:
            defaults[key] = value
    import numpy as np

    params = {'demography': np.array([pop.N]*defaults['simlen'],
              dtype=np.uint32),
              'nregions': [fwdpy11.Region(defaults['beg'],
                           defaults['end'], 1.0)],
              'recregions': [fwdpy11.Region(defaults['beg'],
                             defaults['end'], 1.0)],
              'rates': ((defaults['pneutral']*defaults['theta'])/(4.0*pop.N),
                        ((1.0-defaults['pneutral'])*defaults['theta']) /
                        (4.0*pop.N),
                        defaults['rho']/(4.0*float(pop.N))),
              'gvalue': fwdpy11.Multiplicative(2.0)
              }
    if defaults['dfe'] is None:
        params['sregions'] = []
    else:
        params['sregions'] = [defaults['dfe']]
    return params



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# TYPETEST

import fwdpy11
import unittest
import pickle


class testDiploidCreate(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.d = fwdpy11.DiploidGenotype(0, 0)

    def testPickle(self):
        x = pickle.dumps(self.d)
        dd = pickle.loads(x)
        self.assertEqual(dd.first, 0)
        self.assertEqual(dd.second, 0)


class testDiploidPopulationCreate(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.mutations = fwdpy11.MutationVector()
        self.fixations = fwdpy11.MutationVector()
        self.haploid_genomes = fwdpy11.HaploidGenomeVector()
        self.diploids = fwdpy11.DiploidVector()
        self.mutations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.fixations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.haploid_genomes.append(fwdpy11.HaploidGenome(
            (2, [], [0])))
        self.diploids.append(fwdpy11.DiploidGenotype(0, 0))

    def testConstruction(self):
        pop = fwdpy11.DiploidPopulation(self.diploids, self.haploid_genomes, self.mutations)
        self.assertTrue(pop.N, 1)

    def testStaticMethod(self):
        pop = fwdpy11.DiploidPopulation.create(
            self.diploids, self.haploid_genomes, self.mutations)
        self.assertTrue(type(pop) is fwdpy11.DiploidPopulation)
        # Test that data were moved and not copied:
        self.assertEqual(len(self.diploids), 0)
        self.assertEqual(len(self.haploid_genomes), 0)
        self.assertEqual(len(self.mutations), 0)

    def testStaticMethodWithFixations(self):
        # NOTE: this test kinda stinks,
        # as the "move" requirement is a bit silly in hindsight
        ftimes = [1]
        pop = fwdpy11.DiploidPopulation.create(self.diploids,
                                       self.haploid_genomes,
                                       self.mutations,
                                       self.fixations,
                                       ftimes, 2)
        self.assertTrue(type(pop) is fwdpy11.DiploidPopulation)
        self.assertEqual(len(pop.fixations), len(pop.fixation_times))
        # Test that data were moved and not copied:
        self.assertEqual(len(self.diploids), 0)
        self.assertEqual(len(self.haploid_genomes), 0)
        self.assertEqual(len(self.mutations), 0)
        self.assertEqual(len(self.fixations), 0)
        self.assertEqual(len(ftimes), 1)


class testHaploidGenomeKeySorting(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.mutations = fwdpy11.MutationVector()
        self.fixations = fwdpy11.MutationVector()
        self.haploid_genomes = fwdpy11.HaploidGenomeVector()
        self.diploids = fwdpy11.DiploidVector()
        self.mutations.append(fwdpy11.Mutation(0.1, -0.01, 1.0, 0, 0))
        self.mutations.append(fwdpy11.Mutation(0.01, -2, 1, 0, 0))
        self.haploid_genomes.append(fwdpy11.HaploidGenome(
            (1, [], [0])))
        self.haploid_genomes.append(fwdpy11.HaploidGenome(
            (1, [], [0, 1])))
        self.diploids.append(fwdpy11.DiploidGenotype(0, 1))

    def testUnsortedHaploidGenomes(self):
        with self.assertRaises(ValueError):
            pop = fwdpy11.DiploidPopulation.create(
                self.diploids, self.haploid_genomes, self.mutations)

    def testSortingHaploidGenomes(self):
        fwdpy11.sort_gamete_keys(self.haploid_genomes, self.mutations)
        pop = fwdpy11.DiploidPopulation.create(
            self.diploids, self.haploid_genomes, self.mutations)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import fwdpy11


class test_stopping_criterion_DiploidPopulation(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(1000, 1.0)
        p = {'nregions': [],  # No neutral mutations -- add them later!
             'gvalue': fwdpy11.Additive(2.0),
             'sregions': [fwdpy11.ExpS(0, 1, 1, -0.1)],
             'recregions': [fwdpy11.Region(0, 1, 1)],
             'rates': (0.0, 1e-3, 1e-3),
             # Keep mutations at frequency 1 in the pop if they affect fitness.
             'prune_selected': False,
             'demography':  np.array([1000]*10000, dtype=np.uint32)
             }
        self.params = fwdpy11.ModelParams(**p)

    def test_stop(self):
        rng = fwdpy11.GSLrng(42)

        def generation(pop, simplified):
            if pop.generation >= 50:
                return True
            return False

        fwdpy11.evolvets(
            rng, self.pop, self.params, 100,
            stopping_criterion=generation)
        self.assertEqual(self.pop.generation, 50)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import fwdpy11
import numpy as np

# NOTE: msprime is imported on a per-test basis so as not
# to confict w/other tests of the GSL error handling machinery


class TestConversion(unittest.TestCase):
    @classmethod
    def setUp(self):
        import msprime
        self.ts = msprime.simulate(10, recombination_rate=0.025, Ne=1000)

    # def testGetTablesDiscretizeTime(self):
    #     n, e, ntips, l = fwdpy11.ts_from_msprime._convert_tables(self.ts)
    #     self.assertEqual(ntips, 10)
    #     self.assertEqual(l, 1.0)
    #     na = np.array(n, copy=False)
    #     tzero = np.where(na['time'] == 0.0)
    #     self.assertTrue(len(tzero[0]) == 10)

    def testCreateDiploidPopulation(self):
        pop = fwdpy11.DiploidPopulation.create_from_tskit(self.ts)
        self.assertEqual(pop.N, 5)
        self.assertEqual(pop.tables.genome_length, 1.0)
        md = np.array(pop.diploid_metadata, copy=False)
        n = md['nodes'].flatten()
        self.assertTrue(np.array_equal(
            n, np.arange(2*pop.N, dtype=n.dtype)))


class TestConversionFromMultipleDemes(unittest.TestCase):
    def test_deme_field_of_metadata(self):
        import msprime
        nodes_per_deme = 500
        Ne = 3*nodes_per_deme//2
        Nr = 50.
        config = [msprime.PopulationConfiguration(nodes_per_deme),
                  msprime.PopulationConfiguration(nodes_per_deme),
                  msprime.PopulationConfiguration(nodes_per_deme)]

        events = [msprime.MassMigration(1*Ne, 1, 0, 1.0),
                  msprime.MassMigration(1.5*Ne, 2, 0, 1.)]

        ts = msprime.simulate(population_configurations=config,
                              Ne=Ne,
                              random_seed=98765,
                              recombination_rate=Nr/Ne,
                              demographic_events=events)
        pop = fwdpy11.DiploidPopulation.create_from_tskit(ts)
        self.assertEqual(pop.N,  750)
        alive_nodes = pop.alive_nodes
        self.assertEqual(len(alive_nodes), 2*pop.N)
        for i in range(3):
            for j in alive_nodes[i*nodes_per_deme:(i+1)*nodes_per_deme]:
                self.assertEqual(pop.tables.nodes[j].deme, i)
            k, el = i*nodes_per_deme//2, (i+1)*nodes_per_deme//2
            for j in pop.diploid_metadata[k:el]:
                self.assertEqual(j.deme, i)


if __name__ == "__main__":
    unittest.main()



import fwdpy11
import unittest
import numpy as np


class TestConstruction(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.a = fwdpy11.Additive(2)
        self.list = [fwdpy11.Additive(2), fwdpy11.Additive(2)]

    def test_construct_from_object(self):
        from fwdpy11._fwdpy11 import _dgvalue_pointer_vector
        _dgvalue_pointer_vector(self.a)

    def test_construct_from_list(self):
        from fwdpy11._fwdpy11 import _dgvalue_pointer_vector
        _dgvalue_pointer_vector(self.list)


class TestBadInput(unittest.TestCase):
    def test_construct_from_empty_list(self):
        from fwdpy11._fwdpy11 import _dgvalue_pointer_vector
        with self.assertRaises(ValueError):
            _dgvalue_pointer_vector([])

    def test_invalid_dimensionality(self):
        a = fwdpy11.Additive(2.)
        mvgss = fwdpy11.MultivariateGSS(np.zeros(2), 10)
        mv = fwdpy11.StrictAdditiveMultivariateEffects(2, 0, mvgss)
        from fwdpy11._fwdpy11 import _dgvalue_pointer_vector
        with self.assertRaises(ValueError):
            _dgvalue_pointer_vector([a, mv])

    def test_invalid_types(self):
        from fwdpy11._fwdpy11 import _dgvalue_pointer_vector
        with self.assertRaises(RuntimeError):
            _dgvalue_pointer_vector([1, 2])


if __name__ == "__main__":
    unittest.main()



import unittest
import pickling_composed_classes as PC


class testPickleComposedClass(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.D = PC.Derived()
        self.C = PC.Composed(2.0, self.D)

    def testPickle(self):
        import pickle
        p = pickle.dumps(self.C, -1)
        up = pickle.loads(p)
        self.assertEqual(type(up), type(self.C))
        self.assertEqual(type(up.b), type(self.C.b))


if __name__ == "__main__":
    unittest.main()



"""
"""
import pickle
import lzma
import sys
import numpy as np
import seaborn as sns
import fwdpy11

with lzma.open(sys.argv[1], 'rb') as f:
    pop = pickle.load(f)

genetic_trait_values_from_sim = []
genetic_values_from_ts = []
idx = 0
for n, s, m in pop.sample_timepoints():
    vi = fwdpy11.VariantIterator(pop.tables, s)
    sum_esizes = np.zeros(len(s))
    for variant in vi:
        g = variant.genotypes
        r = variant.records[0]
        mutant = np.where(g == 1)[0]
        sum_esizes[mutant] += pop.mutations[r.key].s
    ind = int(len(s)/2)
    temp_gvalues = np.zeros(ind)
    temp_gvalues += sum_esizes[0::2]
    temp_gvalues += sum_esizes[1::2]
    genetic_values_from_ts.extend(temp_gvalues.tolist())
    genetic_trait_values_from_sim.extend(m['g'].tolist())

# plot the time vs trait value
xyplot = sns.scatterplot(genetic_trait_values_from_sim, genetic_values_from_ts)
xyplot.set_xlabel("From simulation")
xyplot.set_ylabel("From tree sequences")
xyplot.get_figure().savefig("compare_genetic_values.png",
                            dpi=300)



#
# Copyright (C) 2019 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import fwdpy11
import argparse
import sys
import datetime
import numpy as np
import pandas as pd
import seaborn as sns
import concurrent.futures
from collections import namedtuple

THETA = 0.5e3
NSAM_PER_DEME = 10
N_PER_DEME = 1000

Datum = namedtuple('Datum', ['migrate', 'Fst'])


def make_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--nreps', type=int, default=100,
                        help="Number of replicates to simulate per migration rate")
    parser.add_argument('--filename', type=str, default='migtest.png',
                        help="File name for final plot")
    parser.add_argument('--seed', type=int, default=666,
                        help="Random number seed, used for both fwdpy11 and for numpy")
    parser.add_argument('--quickrun', default=False, action='store_true')
    return parser


def plot_results(df, filename, nreps):
    """
    Jitter plot of the raw data + mean and expected values
    """
    # df.sort_values(by=['migrate'])

    df['migrate_str'] = ["{:.2E}".format(i) for i in df['migrate']]
    df.sort_values(by=['migrate'])

    g = df.groupby(['migrate']).mean().reset_index()
    g['efst'] = 1./(1+4*2*N_PER_DEME*g['migrate'])
    g['migrate_str'] = ["{:.2E}".format(i) for i in g['migrate']]

    order = [(i, j) for i, j in zip(g['migrate'], g['migrate_str'])]
    order = sorted(order, key=lambda x: x[0])
    order = [i[1] for i in order]

    fourNm = [(i, "{:.2E}".format(4*N_PER_DEME*i)) for i in g['migrate']]
    fourNm = sorted(fourNm, key=lambda x: x[0])
    fourNm = [i[1] for i in fourNm]

    ax = sns.stripplot(x='migrate_str', y='Fst', alpha=0.5,
                       data=df, jitter=True, order=order)
    now = datetime.datetime.now()
    day = now.strftime("%D")
    title = "Run on {}, based on {} replicates per migration rate.".format(
        day, nreps)
    ax.set_ylim(0, 1)
    ax.set_xlabel(r'$4Nm$')
    ax.set_ylabel(r'$F_{st}$')
    ax.plot('migrate_str', 'Fst', data=g, label=r'mean $F_{st}$')
    ax.plot('migrate_str', 'efst', data=g, label=r'expected $F_{st}$')
    ax.set_xticklabels(fourNm)
    ax.legend()
    ax.set_title(title, fontdict={'fontsize': 10})
    ax.get_figure().savefig(filename)


def fst(m):
    """
    Lazy implementation of Hudson, Slatkin, and
    Maddison's Fst.

    Hudson, R. R., M. Slatkin, and W. P. Maddison. 1992.
    “Estimation of Levels of Gene Flow from DNA Sequence Data.”
    Genetics 132 (2): 583–89.
    """
    nc = 2*NSAM_PER_DEME

    # pi w/in each deme
    dcounts0 = np.sum(m[:, :nc], axis=1)
    dcounts1 = np.sum(m[:, nc:], axis=1)
    pi0 = np.sum(dcounts0*(nc-dcounts0))
    pi0 *= 2.0/(nc*(nc-1))
    pi1 = np.sum(dcounts1*(nc-dcounts1))
    pi1 *= 2.0/(nc*(nc-1))

    # pi in total sample
    dcounts = np.sum(m, axis=1)
    piT = np.sum(dcounts*(m.shape[1]-dcounts))
    piT *= 2./(m.shape[1]*(m.shape[1]-1))

    # Fst
    piS = (pi0+pi1)/2.
    piD = 2.*(piT-piS)
    return piD/(piD+piS)


def runsim(seed, mij):
    migmatrix = np.identity(2)
    migmatrix[:] = mij
    np.fill_diagonal(migmatrix, 0)
    rs = np.sum(migmatrix, axis=1)
    np.fill_diagonal(migmatrix, 1.-rs)
    np.random.seed(seed)
    rng = fwdpy11.GSLrng(seed)
    pop = fwdpy11.DiploidPopulation(2*N_PER_DEME, 1.)
    mm = [fwdpy11.move_individuals(0, 0, 1, 0.5)]

    dd = fwdpy11.DiscreteDemography(mass_migrations=mm,
                                    migmatrix=fwdpy11.MigrationMatrix(migmatrix))

    pdict = {'sregions': [],
             'nregions': [],
             'recregions': [fwdpy11.PoissonInterval(0, 1, 1e2/(4*N_PER_DEME))],
             'gvalue': fwdpy11.Multiplicative(2.),
             'rates': (0, 0, None),
             'demography': dd,
             'simlen': 20*N_PER_DEME
             }

    params = fwdpy11.ModelParams(**pdict)

    fwdpy11.evolvets(rng, pop, params, 100,
                     suppress_table_indexing=True)
    fwdpy11.infinite_sites(rng, pop, THETA/float(4*N_PER_DEME))
    md = np.array(pop.diploid_metadata, copy=False)
    demes = np.unique(md['deme'])
    w = np.where(md['deme'] == demes[0])[0]
    s0 = np.random.choice(w, NSAM_PER_DEME, replace=False)
    w = np.where(md['deme'] == demes[1])[0]
    s1 = np.random.choice(w, NSAM_PER_DEME, replace=False)
    samples = np.concatenate((md['nodes'][s0].flatten(),
                              md['nodes'][s1].flatten()))
    dm = fwdpy11.data_matrix_from_tables(pop.tables, samples, True, False)
    ngm = np.array(dm.neutral, copy=False)
    return Datum(mij, fst(ngm))


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args(sys.argv[1:])
    if args.quickrun is True:
        N_PER_DEME = 200
        max_workers = 1
    else:
        max_workers = None

    rng = fwdpy11.GSLrng(args.seed)
    np.random.seed(args.seed)
    migrates = np.linspace(0, 2.0/(4*N_PER_DEME), 6)
    migrates[0] = 1e-5
    mij = [*migrates.tolist()]*args.nreps
    seeds = np.random.randint(0, np.iinfo(
        np.uint32).max, len(mij)*args.nreps, dtype=np.uint32)
    results = []
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as pool:
        futures = {pool.submit(runsim, i, j) for i, j in zip(seeds, mij)}
        for fut in concurrent.futures.as_completed(futures):
            result = fut.result()
            results.append(result)

    df = pd.DataFrame(results, columns=Datum._fields)
    plot_results(df, args.filename, args.nreps)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


class ModelParams(object):
    """
    Class to hold simulation parameters

    .. versionadded:: 0.1.1

    .. versionchanged:: 0.2.0
        Changed this from a horrible class hierarchy
        into a much simpler, single class.

    .. versionchanged:: 0.6.0

        Updated to support :class:`fwdpy11.DiscreteDemography`

    """

    def __init__(self, **kwargs):
        self.__nregions = []
        self.__sregions = None
        self.__recregions = None
        self.__demography = None
        self.__prune_selected = True
        self.__rates = None
        self.__gvalue = None
        self.__pself = 0.0
        self.__simlen = None
        self.__popsizes = None  # FIXME: this is a hack in 0.6.0
        for key, value in kwargs.items():
            if key in dir(self) and key[:1] != "_":
                setattr(self, key, value)
            elif key not in dir(self):
                raise ValueError(key, " not a valid parameter for this class.")

    def _set_demography(self, value):
        from fwdpy11 import DiscreteDemography
        if isinstance(value, DiscreteDemography):
            return value

        # Assume value is a numpy array of pop
        # sizes over time
        self.simlen = len(value)
        self.__popsizes = value  # FIXME: this is a hack in 0.6.0
        return DiscreteDemography(value)

    @property
    def nregions(self):
        """
        Get or set the neutral regions.
        """
        return self.__nregions

    @nregions.setter
    def nregions(self, nregions_):
        self.__nregions = nregions_

    @property
    def sregions(self):
        """
        Get or set the selected regions.
        """
        return self.__sregions

    @sregions.setter
    def sregions(self, sregions_):
        self.__sregions = sregions_

    @property
    def recregions(self):
        """
        Get or set the recombination regions.
        """
        return self.__recregions

    @recregions.setter
    def recregions(self, recregions_):
        self.__recregions = recregions_

    @property
    def prune_selected(self):
        """
        Get or set whether or not selected fixations
        are to be removed during simulaton.

        When setting, a bool is required.
        """
        return self.__prune_selected

    @prune_selected.setter
    def prune_selected(self, value):
        self.__prune_selected = bool(value)

    @nregions.setter
    def nregions(self, nregions):
        self.__nregions = nregions

    @sregions.setter
    def sregions(self, sregions):
        self.__sregions = sregions

    @property
    def demography(self):
        """
        Get or set demographic history.
        """
        return self.__demography

    @demography.setter
    def demography(self, value):
        self.__demography = self._set_demography(value)

    @property
    def simlen(self):
        """
        Get or set the length of the simulation.

        .. versionadded:: 0.6.0
        """
        return self.__simlen

    @simlen.setter
    def simlen(self, value):
        self.__simlen = value

    @property
    def gvalue(self):
        """
        Get/set the type name of the genetic value
        calculator.
        """
        return self.__gvalue

    @gvalue.setter
    def gvalue(self, gvalue_):
        self.__gvalue = gvalue_

    @property
    def rates(self):
        """
        Get/set mutation and recombination rates.
        """
        return self.__rates

    @rates.setter
    def rates(self, rates_):
        if len(rates_) != 3:
            raise ValueError("length of rates must be 3")
        self.__rates = rates_

    @property
    def mutrate_n(self):
        """
        Return neutral mutation rate(s).
        """
        return self.__rates[0]

    @property
    def mutrate_s(self):
        """
        Return selected mutation rate(s).
        """
        return self.__rates[1]

    @property
    def recrate(self):
        """
        Return recombination rate(s).
        """
        return self.__rates[2]

    @property
    def pself(self):
        """
        Get/set selfing probability/probabilities.
        """
        return self.__pself

    @pself.setter
    def pself(self, pself_):
        self.__pself = pself_

    @property
    def popsizes(self):
        """
        This is a hack in 0.6.0 so that
        simulations w/o tree sequences use the
        same API as previous versions.
        """
        # FIXME: hack introduced in 0.6.0
        return self.__popsizes

    def validate(self):
        """
        Error check model params.

        :raises TypeError: Throws TypeError if validation fails.

        """
        if self.nregions is None:
            raise TypeError("neutral regions cannot be None")
        if self.sregions is None:
            raise TypeError("selected regions cannot be None")
        if self.recregions is None:
            raise TypeError("recombination regions cannot be None")
        if self.demography is None:
            raise TypeError("demography cannot be None")
        if self.prune_selected is None:
            raise TypeError("prune_selected cannot be None")
        if self.gvalue is None:
            raise TypeError("gvalue cannot be None")
        if self.rates is None:
            raise TypeError("rates cannot be None")
        if self.rates[1] > 0 and len(self.sregions) == 0:
            raise ValueError(
                "mutation rate to selected variants is > 0 but no Sregions are defined")
        if self.rates[0] > 0 and len(self.nregions) == 0:
            raise ValueError(
                "mutation rate to neutral variants is > 0 but no Regions are defined")
        if self.simlen is None or self.simlen < 0:
            raise ValueError("simlen cannot be none or < 0")
        # if self.rates[2] > 0 and len(self.recregions) == 0:
        #     raise ValueError("recombination rate is > 0 but no Regions are defined")



import unittest
import fwdpy11
from quick_pops import quick_nonneutral_slocus


class test_mutation_labels(unittest.TestCase):
    def test_label_selected_mutations(self):
        pop = quick_nonneutral_slocus(
            dfe=fwdpy11.ExpS(0, 1, 1, -0.1, 1.0, label=2))
        selected = 0
        for i in pop.mutations:
            if i.neutral is True:
                self.assertEqual(i.label, 0)
            else:
                self.assertEqual(i.label, 2)
                selected += 1
        self.assertTrue(selected > 0) #Else test is useless


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import fwdpy11
import numpy as np
import copy
import os
import pickle
from collections import namedtuple


class Recorder(object):
    def __init__(self, seed, samplesize, timepoints):
        np.random.seed(seed)
        self.samplesize = samplesize
        self.timepoints = timepoints
        self.data = []

    def __call__(self, pop, recorder):
        if len(self.timepoints) > 0:
            if self.timepoints[0] == pop.generation:
                s = np.random.choice(
                    pop.N, self.samplesize, replace=False)
                md = [copy.deepcopy(pop.diploid_metadata[i])
                      for i in s]
                self.data.append((pop.generation, md))
                recorder.assign(s)
                self.timepoints.pop(0)


def set_up_quant_trait_model(simlen=1.0):
    # TODO add neutral variants
    N = 1000
    demography = fwdpy11.DiscreteDemography()
    rho = 2.
    # theta = 100.
    # nreps = 500
    # mu = theta/(4*N)
    r = rho/(4*N)

    GSSmo = fwdpy11.GSSmo([(0, 0, 1), (N, 1, 1)])
    a = fwdpy11.Additive(2.0, GSSmo)
    p = {'nregions': [],
         'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
         'recregions': [fwdpy11.Region(0, 1, 1)],
         'rates': (0.0, 0.025, r),
         'gvalue': a,
         'prune_selected': False,
         'demography': demography,
         'simlen': np.rint(simlen*N).astype(int)
         }
    params = fwdpy11.ModelParams(**p)
    rng = fwdpy11.GSLrng(101*45*110*210)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    return params, rng, pop


def set_up_standard_pop_gen_model(simlen=1.0):
    """
    For this sort of model, when mutations fix, they are
    removed from the simulation, INCLUDING THE TREE
    SEQUENCES. The fact of their existence gets
    recorded in pop.fixations and pop.fixation_times
    """
    # TODO add neutral variants
    N = 1000
    demography = fwdpy11.DiscreteDemography()
    rho = 1.
    # theta = 100.
    # nreps = 500
    # mu = theta/(4*N)
    r = rho/(4*N)

    a = fwdpy11.Multiplicative(2.0)
    pselected = 1e-3
    p = {'nregions': [],
         'sregions': [fwdpy11.GammaS(0, 1, 1.-pselected, mean=-5, shape=1, scaling=2*N),
                      fwdpy11.ConstantS(0, 1, pselected, 1000, scaling=2*N)],
         'recregions': [fwdpy11.Region(0, 1, 1)],
         'rates': (0.0, 0.001, r),
         'gvalue': a,
         'prune_selected': True,
         'demography': demography,
         'simlen': np.rint(simlen*N).astype(int)
         }
    params = fwdpy11.ModelParams(**p)
    rng = fwdpy11.GSLrng(666**2)
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    return params, rng, pop


def mcounts_comparison_details(pop, counts, ts):
    for t in ts.trees():
        for m in t.mutations():
            pos = m.position
            # retrieve this mutation from the fwdpy11 table
            mr = None
            for i in pop.tables.mutations:
                if pop.tables.sites[i.site].position == pos:
                    mr = i
                    break
            assert pop.mutations[mr.key].pos == pos
            tskit_samples = [i for i in t.samples(m.node)]
            if len(tskit_samples) != counts[mr.key]:
                return False
    return True


def mcounts_comparison(pop, ts):
    if len(pop.tables.preserved_nodes) > 0:
        ts2 = ts.simplify(pop.tables.preserved_nodes)
        c = mcounts_comparison_details(pop, pop.mcounts_ancient_samples, ts2)
        if c is not True:
            return c

    ts2 = ts.simplify([i for i in range(2*pop.N)])
    c = mcounts_comparison_details(pop, pop.mcounts, ts2)
    if c is not True:
        return c

    return True


def validate_mut_lookup_content(pop):
    ml = pop.mut_lookup
    ml_contents = []
    for i in ml:
        for j in ml[i]:
            ml_contents.append((i, j))
    ml_contents = sorted(ml_contents)

    mt_contents = []
    for mr in pop.tables.mutations:
        mt_contents.append((pop.tables.sites[mr.site].position, mr.key))
    mt_contents = sorted(mt_contents)
    return ml_contents == mt_contents


class TestBadGenomeLength(unittest.TestCase):
    def test_length_zero(self):
        with self.assertRaises(ValueError):
            fwdpy11.DiploidPopulation(100, 0.)

    def test_negative_length(self):
        with self.assertRaises(ValueError):
            fwdpy11.DiploidPopulation(100, -1.)


class TestTreeSequencesNoAncientSamplesKeepFixations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model(3.0)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        assert max(self.pop.mcounts) == 2 * \
            self.pop.N, "Nothing fixed, so test case is not helpful"

    def test_mut_lookup(self):
        self.assertEqual(len(self.pop.mut_lookup),
                         len(self.pop.tables.mutations))
        self.assertEqual(len(self.pop.mut_lookup), len(self.pop.tables.sites))
        self.assertTrue(validate_mut_lookup_content(self.pop))

    def test_simplify(self):
        tables, idmap = fwdpy11.simplify(self.pop, [i for i in range(10)])
        for i in range(10):
            self.assertTrue(idmap[i] != fwdpy11.NULL_NODE)

    def test_simplify_tables(self):
        tables, idmap = fwdpy11.simplify_tables(
            self.pop.tables, [i for i in range(10)])
        for i in range(10):
            self.assertTrue(idmap[i] != fwdpy11.NULL_NODE)

    def test_simplify_numpy_array(self):
        tables, idmap = fwdpy11.simplify(
            self.pop, np.array([i for i in range(10)]))
        for i in range(10):
            self.assertTrue(idmap[i] != fwdpy11.NULL_NODE)

    def test_simplify_tables_numpy_array(self):
        tables, idmap = fwdpy11.simplify_tables(
            self.pop.tables, np.array([i for i in range(10)]))
        for i in range(10):
            self.assertTrue(idmap[i] != fwdpy11.NULL_NODE)

    def test_dump_to_tskit(self):
        import tskit
        dumped_ts = self.pop.dump_tables_to_tskit()
        self.assertEqual(len(dumped_ts.tables.nodes),
                         len(self.pop.tables.nodes))
        self.assertEqual(len(dumped_ts.tables.edges),
                         len(self.pop.tables.edges))
        self.assertEqual(len(dumped_ts.tables.mutations),
                         len(self.pop.tables.mutations))
        eview = np.array(self.pop.tables.edges, copy=False)
        self.assertEqual(eview['parent'].sum(),
                         dumped_ts.tables.edges.parent.sum())
        self.assertEqual(eview['child'].sum(),
                         dumped_ts.tables.edges.child.sum())
        self.assertEqual(eview['left'].sum(),
                         dumped_ts.tables.edges.left.sum())
        self.assertEqual(eview['right'].sum(),
                         dumped_ts.tables.edges.right.sum())
        tv = fwdpy11.TreeIterator(self.pop.tables,
                                  [i for i in range(2*self.pop.N)])
        tt_fwd = 0
        for t in tv:
            tt_fwd += t.total_time(self.pop.tables.nodes)
        tt_tskit = 0
        for t in dumped_ts.trees():
            tt_tskit += t.get_total_branch_length()
        self.assertEqual(tt_fwd, tt_tskit)

        # Now, we make sure that the metadata can
        # be decoded
        md = tskit.unpack_bytes(dumped_ts.tables.individuals.metadata,
                                dumped_ts.tables.individuals.metadata_offset)
        for i, j in zip(self.pop.diploid_metadata, md):
            d = eval(j)
            self.assertEqual(i.g, d['g'])
            self.assertEqual(i.w, d['w'])
            self.assertEqual(i.e, d['e'])
            self.assertEqual(i.label, d['label'])
            self.assertEqual(i.parents, d['parents'])
            self.assertEqual(i.sex, d['sex'])
            self.assertEqual(i.deme, d['deme'])
            self.assertEqual(i.geography, d['geography'])

        # Test that we can go backwards from node table to individuals
        samples = np.where(dumped_ts.tables.nodes.flags ==
                           tskit.NODE_IS_SAMPLE)[0]
        self.assertEqual(len(samples), 2*self.pop.N)
        for i in samples[::2]:
            ind = i//2
            d = eval(md[ind])
            fwdpy11_md = self.pop.diploid_metadata[ind]
            self.assertEqual(fwdpy11_md.g, d['g'])
            self.assertEqual(fwdpy11_md.w, d['w'])
            self.assertEqual(fwdpy11_md.e, d['e'])
            self.assertEqual(fwdpy11_md.label, d['label'])
            self.assertEqual(fwdpy11_md.parents, d['parents'])
            self.assertEqual(fwdpy11_md.sex, d['sex'])
            self.assertEqual(fwdpy11_md.deme, d['deme'])
            self.assertEqual(fwdpy11_md.geography, d['geography'])

        md = tskit.unpack_bytes(dumped_ts.tables.mutations.metadata,
                                dumped_ts.tables.mutations.metadata_offset)
        for i, j, k in zip(self.pop.tables.mutations,
                           dumped_ts.tables.mutations.site,
                           md):
            d = eval(k)
            self.assertEqual(i.key, d['key'])
            site = dumped_ts.tables.sites[j]
            m = self.pop.mutations[d['key']]
            self.assertEqual(site.position, m.pos)
            self.assertEqual(d['s'], m.s)
            self.assertEqual(d['h'], m.h)
            self.assertTrue(np.array_equal(np.array(d['esizes']), m.esizes))
            self.assertTrue(np.array_equal(
                np.array(d['heffects']), m.heffects))
            self.assertEqual(d['label'], m.label)
            self.assertEqual(d['neutral'], m.neutral)

        self.assertEqual(mcounts_comparison(self.pop, dumped_ts), True)

    def test_TreeIterator(self):
        # The first test ensures that TreeIterator
        # simply holds a reference to the input tables,
        # rather than a (deep) copy, which would have
        # a different address
        tv = fwdpy11.TreeIterator(self.pop.tables,
                                  [i for i in range(2*self.pop.N)],
                                  True, 0, 1)
        self.assertTrue(tv.tables is self.pop.tables)

        with self.assertRaises(ValueError):
            tv = fwdpy11.TreeIterator(self.pop.tables,
                                      [i for i in range(2*self.pop.N)],
                                      True, 1, 0)
        with self.assertRaises(ValueError):
            tv = fwdpy11.TreeIterator(self.pop.tables,
                                      [i for i in range(2*self.pop.N)],
                                      False, 1, 0)

        for i in np.arange(0., 1., 0.1):
            tv = fwdpy11.TreeIterator(self.pop.tables,
                                      [i for i in range(2*self.pop.N)],
                                      True, i, i+0.1)
            for ti in tv:
                a = ti.left < i+0.1
                b = i < ti.right
                self.assertTrue(a and b)

            tv = fwdpy11.TreeIterator(self.pop.tables,
                                      [i for i in range(2*self.pop.N)],
                                      False, i, i+0.1)
            for ti in tv:
                a = ti.left < i+0.1
                b = i < ti.right
                self.assertTrue(a and b)

    def test_TreeIterator_iterate_sites(self):
        # TODO: need test of empty tree sequence
        # and tree sequence where mutations aren't
        # on every tree
        tv = fwdpy11.TreeIterator(self.pop.tables,
                                  [i for i in range(2*self.pop.N)])
        nsites_visited = 0
        for tree in tv:
            for s in tree.sites():
                self.assertTrue(s.position >= tree.left)
                self.assertTrue(s.position < tree.right)
                nsites_visited += 1
        self.assertEqual(len(self.pop.tables.sites), nsites_visited)

        site_table = np.array(self.pop.tables.sites, copy=False)
        for i in np.arange(0., 1., 0.1):
            tv = fwdpy11.TreeIterator(self.pop.tables,
                                      [i for i in range(2*self.pop.N)], False, i, i+0.1)
            nsites_visited = 0
            idx = np.where((site_table['position'] >= i) & (
                site_table['position'] < i+0.1))[0]
            nsites_in_interval = len(idx)
            for tree in tv:
                for s in tree.sites():
                    self.assertTrue(s.position >= tree.left)
                    self.assertTrue(s.position < tree.right)
                    nsites_visited += 1
            self.assertEqual(nsites_visited, nsites_in_interval)

    def test_TreeIterator_iterate_mutations(self):
        tv = fwdpy11.TreeIterator(self.pop.tables,
                                  [i for i in range(2*self.pop.N)])
        sites = np.array(self.pop.tables.sites, copy=False)
        nsites_visited = 0
        for tree in tv:
            for m in tree.mutations():
                self.assertTrue(sites['position'][m.site] >= tree.left)
                self.assertTrue(sites['position'][m.site] < tree.right)
                nsites_visited += 1
        self.assertEqual(len(self.pop.tables.sites), nsites_visited)

        for i in np.arange(0., 1., 0.1):
            tv = fwdpy11.TreeIterator(self.pop.tables,
                                      [i for i in range(2*self.pop.N)], False, i, i+0.1)
            nsites_visited = 0
            idx = np.where((sites['position'] >= i) & (
                sites['position'] < i+0.1))[0]
            nsites_in_interval = len(idx)
            for tree in tv:
                for m in tree.mutations():
                    self.assertTrue(sites['position'][m.site] >= tree.left)
                    self.assertTrue(sites['position'][m.site] < tree.right)
                    nsites_visited += 1
            self.assertEqual(nsites_visited, nsites_in_interval)

    def test_leaf_counts_vs_mcounts(self):
        tv = fwdpy11.TreeIterator(self.pop.tables,
                                  [i for i in range(2*self.pop.N)])
        mv = np.array(self.pop.tables.mutations, copy=False)
        muts = self.pop.mutations_ndarray
        p = muts['pos']
        for t in tv:
            l, r = t.left, t.right
            mt = [i for i in mv if p[i[1]] >= l and p[i[1]] < r]
            for i in mt:
                self.assertEqual(t.leaf_counts(i[0]),
                                 self.pop.mcounts[i[1]])

    def test_simplify_to_sample(self):
        """
        Simplify to a sample using fwdpy11 and
        tskit, then test that total time on output
        is the same from both sources and that
        the mutation tables contain the same
        positions after simplification.
        """
        dumped_ts = self.pop.dump_tables_to_tskit()
        tt = 0.0
        for i in self.pop.tables.nodes:
            tt += i.time
        samples = np.arange(0, 2*self.pop.N, 50, dtype=np.int32)
        mspts = dumped_ts.simplify(samples=samples.tolist())
        fp11ts, idmap = fwdpy11.simplify(self.pop, samples)
        for i in range(len(fp11ts.edges)):
            self.assertTrue(fp11ts.edges[i].parent < len(fp11ts.nodes))
            self.assertTrue(fp11ts.edges[i].child < len(fp11ts.nodes))
        for s in samples:
            self.assertEqual(
                fp11ts.nodes[idmap[s]].time, self.pop.generation)
        tt_fwd = 0.0
        tv = fwdpy11.TreeIterator(fp11ts, [i for i in range(len(samples))])
        for t in tv:
            tt_fwd += t.total_time(fp11ts.nodes)
        tt_tskit = 0.0
        for t in mspts.trees():
            tt_tskit += t.get_total_branch_length()
        self.assertEqual(tt_fwd, tt_tskit)

        self.assertEqual(len(fp11ts.mutations),
                         len(mspts.tables.mutations))
        fp11_pos = np.array([self.pop.mutations[i.key].pos
                             for i in fp11ts.mutations])
        fp11_pos = np.sort(fp11_pos)
        msp_pos = np.sort(mspts.tables.sites.position)
        self.assertTrue(np.array_equal(fp11_pos, msp_pos))

    def test_genotype_matrix(self):
        """
        Make data matrix objects from the tree sequences
        and compare their contents to those of tskit
        as well as to an explicit calculation of mutation counts.
        """
        dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                             [i for i in range(2*self.pop.N)],
                                             False, True, True)
        sa = np.array(dm.selected)
        cs = np.sum(sa, axis=1)
        dumped_ts = self.pop.dump_tables_to_tskit()
        mm = dumped_ts.genotype_matrix()
        mc = np.sum(mm, axis=1)
        ec = np.zeros(len(self.pop.mutations), dtype=np.uint32)
        for d in self.pop.diploids:
            for m in self.pop.haploid_genomes[d.first].smutations:
                ec[m] += 1
            for m in self.pop.haploid_genomes[d.second].smutations:
                ec[m] += 1
        for i, j, k in zip(self.pop.tables.mutations, cs, mc):
            self.assertEqual(self.pop.mcounts[i.key],
                             ec[i.key])
            self.assertEqual(ec[i.key], j)
            self.assertEqual(ec[i.key], k)
        self.assertTrue(np.array_equal(ec, np.array(self.pop.mcounts)))
        self.assertEqual(ec.sum(), cs.sum())
        self.assertEqual(ec.sum(), mc.sum())
        self.assertTrue(np.array_equal(sa, mm))
        self.assertTrue(np.array_equal(cs, mc))

    def test_genotype_matrix_ranges(self):
        dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                             [i for i in range(
                                                 2*self.pop.N)],
                                             False, True)
        spos = np.array(dm.selected.positions)
        for i in np.arange(0, self.pop.tables.genome_length, 0.1):
            dmi = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                                  [i for i in range(
                                                      2*self.pop.N)],
                                                  record_neutral=False,
                                                  record_selected=True,
                                                  begin=i, end=i+0.1)
            w = np.where((spos >= i) & (spos < i+0.1))[0]
            self.assertTrue(np.array_equal(
                spos[w], np.array(dmi.selected.positions)))

    def test_VariantIterator(self):
        """
        Test VariantIterator by asserting
        that sum of genotypes equal values in
        the corresponding DataMatrix and
        those in pop.mcounts
        """
        dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                             [i for i in range(
                                                 2*self.pop.N)],
                                             False, True, True)
        sa = np.array(dm.selected)
        cs = np.sum(sa, axis=1)
        i = 0
        vi = fwdpy11.VariantIterator(self.pop.tables,
                                     [i for i in range(2*self.pop.N)])
        for v in vi:
            c = self.pop.mcounts[self.pop.tables.mutations[i].key]
            self.assertEqual(c, cs[i])
            self.assertEqual(c, v.genotypes.sum())
            i += 1
        mc = np.array(self.pop.mcounts)
        self.assertEqual(i, len(np.where(mc > 0)[0]))
        self.assertEqual(i, len(self.pop.tables.mutations))

    def test_VariantIteratorFromPopulation(self):
        dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                             [i for i in range(
                                                 2*self.pop.N)],
                                             False, True, True)
        sa = np.array(dm.selected)
        cs = np.sum(sa, axis=1)
        i = 0
        vi = fwdpy11.VariantIterator(self.pop)
        for v in vi:
            c = self.pop.mcounts[self.pop.tables.mutations[i].key]
            self.assertEqual(c, cs[i])
            self.assertEqual(c, v.genotypes.sum())
            i += 1
        mc = np.array(self.pop.mcounts)
        self.assertEqual(i, len(np.where(mc > 0)[0]))
        self.assertEqual(i, len(self.pop.tables.mutations))

    def test_VariantIteratorBeginEnd(self):
        for i in np.arange(0, self.pop.tables.genome_length, 0.1):
            vi = fwdpy11.VariantIterator(self.pop.tables,
                                         [i for i in range(2*self.pop.N)], i,
                                         i+0.1)
            nm = len([j for j in self.pop.tables.mutations if self.pop.mutations[j.key].pos >= i and
                      self.pop.mutations[j.key].pos < i+0.1])
            nseen = 0
            for v in vi:
                r = v.records[0]
                self.assertTrue(self.pop.mutations[r.key].pos >= i)
                self.assertTrue(self.pop.mutations[r.key].pos < i+0.1)
                nseen += 1
            self.assertEqual(nm, nseen)

        # test bad start/stop
        with self.assertRaises(ValueError):
            vi = fwdpy11.VariantIterator(self.pop.tables,
                                         [i for i in range(2*self.pop.N)], begin=0.5, end=0.25)

    def test_count_mutations(self):
        mc = fwdpy11.count_mutations(self.pop,
                                     [i for i in range(2*self.pop.N)])
        pmc = np.array(self.pop.mcounts)
        self.assertTrue(np.array_equal(mc, pmc))

    def test_binary_round_trip(self):
        ofile = "poptest_no_ancient_preserve_fixations.bin"
        self.pop.dump_to_file(ofile)
        pop2 = fwdpy11.DiploidPopulation.load_from_file(ofile)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_fast_pickling(self):
        p = pickle.dumps(self.pop, -1)
        up = pickle.loads(p)
        self.assertTrue(self.pop == up)

    def test_slow_pickling_to_file(self):
        ofile = "poptest_no_ancient_preserve_fixations.pickle"
        with open(ofile, 'wb') as f:
            self.pop.pickle_to_file(f)
        with open(ofile, 'rb') as f:
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(f)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)


class TestTreeSequencesWithAncientSamplesKeepFixations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model(3.0)
        self.stimes = [i for i in range(1, 101)]
        self.recorder = Recorder(42, 10, self.stimes)
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 100, self.recorder)
        assert max(self.pop.mcounts) == 2 * \
            self.pop.N, "Nothing fixed, so test case is not helpful"

    def test_mut_lookup(self):
        self.assertEqual(len(self.pop.mut_lookup),
                         len(self.pop.tables.mutations))
        self.assertEqual(len(self.pop.mut_lookup), len(self.pop.tables.sites))
        self.assertTrue(validate_mut_lookup_content(self.pop))

    def test_serialization_via_file_round_trip(self):
        with open("test_ts_serialization_via_file.pickle", "wb") as f:
            self.pop.pickle_to_file(f)

        with open("test_ts_serialization_via_file.pickle", "rb") as f:
            up = fwdpy11.DiploidPopulation.load_from_pickle_file(f)
        self.assertTrue(self.pop == up)
        self.assertTrue(self.pop.mut_lookup == up.mut_lookup)

    def test_pickling_round_trip(self):
        p = pickle.dumps(self.pop, -1)
        up = pickle.loads(p)
        self.assertTrue(self.pop == up)
        self.assertTrue(self.pop.mut_lookup == up.mut_lookup)

    def test_count_mutations_preserved_samples(self):
        mc = fwdpy11.count_mutations(self.pop,
                                     self.pop.tables.preserved_nodes)
        pmc = np.array(self.pop.mcounts_ancient_samples)
        self.assertTrue(np.array_equal(mc, pmc))

    def test_ancient_sample_times(self):
        times = []
        for t, n, md in self.pop.sample_timepoints(False):
            times.append(int(t))
        self.assertEqual(times, [i for i in range(1, 101)])

    def test_VariantIteratorFromPreservedSamples(self):
        n = np.array(self.pop.tables.nodes)
        pn = np.array(self.pop.tables.preserved_nodes)
        at = n['time'][pn]
        for u in np.unique(at):
            n = pn[np.where(at == u)[0]]
            vi = fwdpy11.VariantIterator(self.pop.tables, n)
            for variant in vi:
                k = variant.records[0]
                self.assertNotEqual(k.node, fwdpy11.NULL_NODE)
                self.assertNotEqual(k.key, np.iinfo(np.uint64).max)

    def test_mcounts_from_ancient_samples(self):
        vi = fwdpy11.VariantIterator(
            self.pop.tables, self.pop.tables.preserved_nodes)
        for v in vi:
            k = v.records[0]
            self.assertEqual(
                self.pop.mcounts_ancient_samples[k.key], v.genotypes.sum())

    def test_sample_traverser(self):
        timepoints = [i for i in range(1, 101)]
        amd = np.array(self.pop.ancient_sample_metadata, copy=False)
        n = np.array(self.pop.tables.nodes)
        at = n['time'][amd['nodes'][:, 0]]
        pn = np.array(self.pop.tables.preserved_nodes)
        self.assertEqual(2*len(amd), len(pn))
        # j contains (time, nodes, metadata).  The metadata
        # also contain nodes
        for i, j, d in zip(timepoints,
                           self.pop.sample_timepoints(False), self.recorder.data):
            self.assertEqual(i, d[0])
            self.assertEqual(i, j[0])
            self.assertTrue(all(n['time'][j[1]] == i))
            mdn = j[2]['nodes'].flatten()
            self.assertTrue(np.array_equal(j[1], mdn))
            idx = np.where(at == i)[0]
            self.assertTrue(np.array_equal(j[2], amd[idx]))
            # Compare metadata stored in recorder
            # to that preserved internally by the simulation
            for k, l in zip(idx, d[1]):
                self.assertEqual(amd['g'][k], l.g)
                self.assertEqual(amd['w'][k], l.w)
                self.assertEqual(amd['e'][k], l.e)
                self.assertEqual(amd['label'][k], l.label)
                self.assertEqual(amd['parents'][k][0], l.parents[0])
                self.assertEqual(amd['parents'][k][1], l.parents[1])

            # Extract out the nodes from preserved_nodes
            idx = np.where(n['time'][self.pop.tables.preserved_nodes] == i)[0]
            self.assertTrue(np.array_equal(
                np.array(self.pop.tables.preserved_nodes)[idx], j[1]))

    def test_binary_round_trip(self):
        ofile = "poptest_with_ancient_preserve_fixations.bin"
        self.pop.dump_to_file(ofile)
        pop2 = fwdpy11.DiploidPopulation.load_from_file(ofile)
        self.assertTrue(self.pop == pop2)
        self.assertTrue(self.pop.mut_lookup == pop2.mut_lookup)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_fast_pickling(self):
        p = pickle.dumps(self.pop, -1)
        up = pickle.loads(p)
        self.assertTrue(self.pop == up)

    def test_slow_pickling_to_file(self):
        ofile = "poptest_with_ancient_preserve_fixations.pickle"
        with open(ofile, 'wb') as f:
            self.pop.pickle_to_file(f)
        with open(ofile, 'rb') as f:
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(f)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_dump_to_tskit(self):
        dumped_ts = self.pop.dump_tables_to_tskit()
        self.assertEqual(mcounts_comparison(self.pop, dumped_ts), True)


class TestMutationCounts(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model()
        self.params.demography = np.array([self.pop.N]*200, dtype=np.uint32)
        self.pop2 = copy.deepcopy(self.pop)
        self.rng2 = fwdpy11.GSLrng(101*45*110*210)
        params2 = copy.deepcopy(self.params)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        fwdpy11.evolvets(self.rng2, self.pop2, params2,
                         100, remove_extinct_variants=False)

    def test_mut_lookup(self):
        self.assertEqual(len(self.pop.mut_lookup),
                         len(self.pop.tables.mutations))
        self.assertEqual(len(self.pop.mut_lookup), len(self.pop.tables.sites))
        self.assertTrue(validate_mut_lookup_content(self.pop))

    def test_mutation_counts(self):
        self.assertEqual(sum(self.pop.mcounts), sum(self.pop2.mcounts))

    def test_mutation_counts_details(self):
        mc = [i for i in self.pop2.mcounts if i > 0]
        self.assertTrue(
            all([i == j for i, j in zip(self.pop.mcounts, mc)]) is True)

    def test_mutation_counts_from_genomes(self):
        """
        Removing extinct genomes happens after
        the final mutation count, so let's
        make sure we can reconstruct the counts
        from the remapped genomes
        """
        mc = [0] * len(self.pop.mutations)
        mc2 = [0] * len(self.pop2.mutations)
        for g in self.pop.haploid_genomes:
            if g.n > 0:
                for k in g.smutations:
                    mc[k] += g.n
        for g in self.pop2.haploid_genomes:
            if g.n > 0:
                for k in g.smutations:
                    mc2[k] += g.n
        self.assertTrue(
            all([i == j for i, j in zip(mc, self.pop.mcounts)]) is True)
        self.assertTrue(
            all([i == j for i, j in zip(mc2, self.pop2.mcounts)]) is True)


class TestUpdateTiming(unittest.TestCase):
    """
    Motivated by issue 437.
    This test case constructs a situation
    where we know exactly when a change should
    happen and what its outcome should be.
    If the optimum shifts at generation X,
    then the current parents should experience
    reduced fitness at that time, which we can
    measure by their mean fitness, which is trivial
    if there are no mutations.
    """
    def testTiming(self):
        """
        Make sure that timings occur when we expect.
        """
        from test_tree_sequences import set_up_quant_trait_model
        params, rng, pop = set_up_quant_trait_model(1.1)
        params.rates = (0, 0, 0)

        class Recorder(object):
            def __init__(self):
                self.data = []

            def __call__(self, pop, sampler):
                if pop.generation >= 990:
                    md = np.array(pop.diploid_metadata, copy=False)
                    self.data.append((pop.generation, md['w'].mean()))
        r = Recorder()
        fwdpy11.evolvets(rng, pop, params, 1000, r)
        for i in r.data:
            if i[0] < pop.N:
                self.assertAlmostEqual(i[1], 1.0)
            else:
                self.assertAlmostEqual(i[1], np.exp(-0.50))


class TestTreeSequencesNoAncientSamplesPruneFixations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        fwdpy11.evolvets(self.rng, self.pop, self.params,
                         100, track_mutation_counts=True)
        assert len(
            self.pop.fixations) > 0, "Nothing fixed, so test case is not helpful"

    def test_mut_lookup(self):
        self.assertEqual(len(self.pop.mut_lookup),
                         len(self.pop.tables.mutations))
        self.assertEqual(len(self.pop.mut_lookup), len(self.pop.tables.sites))
        self.assertTrue(validate_mut_lookup_content(self.pop))

    def test_max_mcounts(self):
        self.assertTrue(max(self.pop.mcounts) < 2*self.pop.N)

    def test_mutation_table_contents(self):
        self.assertEqual(len(self.pop.mcounts), len(
            self.pop.mcounts_ancient_samples))
        for m in self.pop.tables.mutations:
            self.assertTrue(m.key < len(self.pop.mutations))
            self.assertTrue(self.pop.mcounts[m.key] < 2*self.pop.N)

        for f in self.pop.fixations:
            m = None
            for i in self.pop.tables.mutations:
                if self.pop.tables.sites[i.site].position == f.pos:
                    m = i
                    break
            if m is not None:
                # Then the mutation position got re-used, and
                # so the mutations must have different origin times
                # NOTE: this probably has not happened!
                self.assertNotEqual(f.g, self.pop.mutations[m.key].g)

    def test_dump_to_tskit(self):
        ts = self.pop.dump_tables_to_tskit()
        self.assertEqual(mcounts_comparison(self.pop, ts), True)

    def test_binary_round_trip(self):
        ofile = "poptest_no_ancient_prune_fixations.bin"
        self.pop.dump_to_file(ofile)
        pop2 = fwdpy11.DiploidPopulation.load_from_file(ofile)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_fast_pickling(self):
        p = pickle.dumps(self.pop, -1)
        up = pickle.loads(p)
        self.assertTrue(self.pop == up)

    def test_slow_pickling_to_file(self):
        ofile = "poptest_no_ancient_prune_fixations.pickle"
        with open(ofile, 'wb') as f:
            self.pop.pickle_to_file(f)
        with open(ofile, 'rb') as f:
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(f)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_genotype_matrix(self):
        dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                             [i for i in range(2*self.pop.N)],
                                             False, True, True)
        rc = np.sum(dm.selected, axis=1)
        index = [i for i in range(len(self.pop.mcounts))]
        index = sorted(index, key=lambda x: self.pop.mutations[x].pos)
        mc = [self.pop.mcounts[i] for i in index if self.pop.mcounts[i] > 0]
        self.assertTrue(np.array_equal(rc, np.array(mc)))


class TestTreeSequencesWithAncientSamplesPruneFixations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        self.stimes = [i for i in range(1, 101)]
        self.recorder = Recorder(42, 10, self.stimes)
        fwdpy11.evolvets(self.rng, self.pop, self.params,
                         100, self.recorder, track_mutation_counts=True)
        assert len(
            self.pop.fixations) > 0, "Nothing fixed, so test case is not helpful"

    def test_mut_lookup(self):
        self.assertEqual(len(self.pop.mut_lookup),
                         len(self.pop.tables.mutations))
        self.assertEqual(len(self.pop.mut_lookup), len(self.pop.tables.sites))
        self.assertTrue(validate_mut_lookup_content(self.pop))

    def test_mutation_table_contents(self):
        self.assertEqual(len(self.pop.mcounts), len(
            self.pop.mcounts_ancient_samples))
        for m in self.pop.tables.mutations:
            self.assertTrue(m.key < len(self.pop.mutations))
            if self.pop.mcounts[m.key] < 2*self.pop.N:
                is_fixed = False
            else:
                is_fixed = True
            if self.pop.mcounts_ancient_samples[m.key] != 0:
                is_found_ancient = True
            else:
                is_found_ancient = False
            self.assertTrue(is_fixed is False or (
                is_fixed is True and is_found_ancient is True))

        for f in self.pop.fixations:
            m = None
            for i in self.pop.tables.mutations:
                if self.pop.tables.sites[i.site].position == f.pos:
                    m = i
                    break
            if m is not None:
                # Then the mutation position got re-used, and
                # so the mutations must have different origin times
                # NOTE: this probably has not happened!
                self.assertNotEqual(f.g, self.pop.mutations[m.key].g)

    def test_dump_to_tskit(self):
        ts = self.pop.dump_tables_to_tskit()
        self.assertEqual(mcounts_comparison(self.pop, ts), True)

    def test_binary_round_trip(self):
        ofile = "poptest_with_ancient_prune_fixations.bin"
        self.pop.dump_to_file(ofile)
        pop2 = fwdpy11.DiploidPopulation.load_from_file(ofile)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_fast_pickling(self):
        p = pickle.dumps(self.pop, -1)
        up = pickle.loads(p)
        self.assertTrue(self.pop == up)

    def test_slow_pickling_to_file(self):
        ofile = "poptest_with_ancient_prune_fixations.pickle"
        with open(ofile, 'wb') as f:
            self.pop.pickle_to_file(f)
        with open(ofile, 'rb') as f:
            pop2 = fwdpy11.DiploidPopulation.load_from_pickle_file(f)
        self.assertTrue(self.pop == pop2)
        if os.path.exists(ofile):
            os.remove(ofile)

    def test_genotype_matrix(self):
        dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                             [i for i in range(2*self.pop.N)],
                                             False, True, True)
        rc = np.sum(dm.selected, axis=1)
        index = [i for i in range(len(self.pop.mcounts))]
        index = sorted(index, key=lambda x: self.pop.mutations[x].pos)
        mc = [self.pop.mcounts[i] for i in index if self.pop.mcounts[i] > 0]
        self.assertTrue(np.array_equal(rc, np.array(mc)))


class TestSimplificationInterval(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.N = 1000
        self.demography = np.array([self.N]*100, dtype=np.uint32)
        self.rho = 1.
        self.theta = 100.
        self.nreps = 500
        self.mu = self.theta/(4*self.N)
        self.r = self.rho/(4*self.N)
        self.GSS = fwdpy11.GSS(VS=1, opt=0)
        a = fwdpy11.Additive(2.0, self.GSS)
        self.p = {'nregions': [],
                  'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
                  'recregions': [fwdpy11.Region(0, 1, 1)],
                  'rates': (0.0, 0.025, self.r),
                  'gvalue': a,
                  'prune_selected': False,
                  'demography': self.demography
                  }
        self.params = fwdpy11.ModelParams(**self.p)
        self.rng = fwdpy11.GSLrng(101*45*110*210)
        self.pop = fwdpy11.DiploidPopulation(self.N, 1.0)
        self.recorder = fwdpy11.RandomAncientSamples(seed=42,
                                                     samplesize=10,
                                                     timepoints=[i for i in range(1, 101)])

    def testEvolve(self):
        # TODO: actually test something here :)
        fwdpy11.evolvets(
            self.rng, self.pop, self.params, 1, self.recorder)
        samples = [i for i in range(2*self.pop.N)] + \
            self.pop.tables.preserved_nodes
        vi = fwdpy11.TreeIterator(self.pop.tables, samples)

    def test_mut_lookup(self):
        self.assertEqual(len(self.pop.mut_lookup),
                         len(self.pop.tables.mutations))
        self.assertEqual(len(self.pop.mut_lookup), len(self.pop.tables.sites))
        self.assertTrue(validate_mut_lookup_content(self.pop))


class TestFixationPreservation(unittest.TestCase):
    def testQtraitSim(self):
        N = 1000
        demography = np.array([N]*3*N, dtype=np.uint32)
        rho = 1.
        r = rho/(4*N)

        GSS = fwdpy11.GSS(VS=1, opt=1)
        a = fwdpy11.Additive(2.0, GSS)
        p = {'nregions': [],
             'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
             'recregions': [fwdpy11.Region(0, 1, 1)],
             'rates': (0.0, 0.005, r),
             'gvalue': a,
             'prune_selected': False,
             'demography': demography
             }
        params = fwdpy11.ModelParams(**p)
        rng = fwdpy11.GSLrng(101*45*110*210)
        pop = fwdpy11.DiploidPopulation(N, 1.0)
        fwdpy11.evolvets(rng, pop, params, 100,
                         track_mutation_counts=True)
        mc = fwdpy11.count_mutations(pop.tables, pop.mutations,
                                     [i for i in range(2*pop.N)])
        assert len(pop.fixations) > 0, "Test is meaningless without fixations"
        fixations = np.where(mc == 2*pop.N)[0]
        self.assertEqual(len(fixations), len(pop.fixations))

        # Brute-force calculation of fixations
        brute_force = np.zeros(len(pop.mutations), dtype=np.int32)
        for g in pop.haploid_genomes:
            if g.n > 0:
                for k in g.smutations:
                    brute_force[k] += g.n

        self.assertTrue(np.array_equal(brute_force, mc))
        self.assertTrue(np.array_equal(brute_force, pop.mcounts))

    def testPopGenSim(self):
        N = 1000
        demography = np.array([N]*2*N, dtype=np.uint32)
        rho = 1.
        r = rho/(4*N)

        a = fwdpy11.Multiplicative(2.0)
        p = {'nregions': [],
             'sregions': [fwdpy11.ExpS(0, 1, 1, 0.01)],
             'recregions': [fwdpy11.Region(0, 1, 1)],
             'rates': (0.0, 0.00005, r),
             'gvalue': a,
             'prune_selected': True,
             'demography': demography
             }
        params = fwdpy11.ModelParams(**p)
        rng = fwdpy11.GSLrng(101*45*110*210)
        pop = fwdpy11.DiploidPopulation(N, 1.0)
        fwdpy11.evolvets(rng, pop, params, 100,
                         track_mutation_counts=True)
        mc = fwdpy11.count_mutations(pop.tables, pop.mutations,
                                     [i for i in range(2*pop.N)])
        assert len(pop.fixations) > 0, "Test is meaningless without fixations"
        fixations = np.where(mc == 2*pop.N)[0]
        self.assertEqual(len(fixations), 0)

        # Brute-force calculation of fixations
        brute_force = np.zeros(len(pop.mutations), dtype=np.int32)
        for g in pop.haploid_genomes:
            if g.n > 0:
                for k in g.smutations:
                    brute_force[k] += g.n

        self.assertTrue(np.array_equal(brute_force, mc))
        self.assertTrue(np.array_equal(brute_force, pop.mcounts))


class TestMetaData(unittest.TestCase):
    """
    Integration test.

    Rebuild genetic values from mutations
    stored in tree sequence
    """

    def testQtraitSim(self):
        N = 1000
        rho = 1.
        r = rho/(4*N)

        GSS = fwdpy11.GSS(VS=1, opt=1)
        a = fwdpy11.Additive(2.0, GSS)
        p = {'nregions': [],
             'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
             'recregions': [fwdpy11.PoissonInterval(0, 1, r)],
             'rates': (0.0, 0.005, None),
             'gvalue': a,
             'prune_selected': False,
             'demography': fwdpy11.DiscreteDemography(),
             'simlen': N
             }
        params = fwdpy11.ModelParams(**p)
        rng = fwdpy11.GSLrng(101*45*110*210)
        pop = fwdpy11.DiploidPopulation(N, 1.0)

        # NOTE: we are trying to track the fate of C++
        # types based on their Python instances. We can
        # run into trouble as the underlying data are
        # over-written.  Thus, we store some of it
        # in a new class for testing.
        MD = namedtuple("MD", ['g', 'e', 'w', 'label'])

        class Recorder(object):
            """ Records entire pop every 100 generations """

            def __init__(self):
                self.data = []

            def __call__(self, pop, recorder):
                if pop.generation % 100 == 0.0:
                    for i in pop.diploid_metadata:
                        self.data.append(
                            (pop.generation, MD(i.g, i.e, i.w, i.label)))
                    recorder.assign(np.arange(pop.N, dtype=np.int32))

        recorder = Recorder()
        fwdpy11.evolvets(rng, pop, params, 100, recorder)

        ancient_sample_metadata = np.array(
            pop.ancient_sample_metadata, copy=False)
        alive_sample_metadata = np.array(pop.diploid_metadata, copy=False)
        metadata = np.hstack((ancient_sample_metadata, alive_sample_metadata))

        nodes = np.array(pop.tables.nodes, copy=False)
        metadata_nodes = metadata['nodes'].flatten()
        metadata_node_times = nodes['time'][metadata_nodes]
        metadata_record_times = nodes['time'][metadata['nodes'][:, 0]]

        genetic_trait_values_from_sim = []
        genetic_values_from_ts = []
        for u in np.unique(metadata_node_times):
            samples_at_time_u = metadata_nodes[np.where(
                metadata_node_times == u)]
            vi = fwdpy11.VariantIterator(
                pop.tables, samples_at_time_u)
            sum_esizes = np.zeros(len(samples_at_time_u))
            for variant in vi:
                g = variant.genotypes
                r = variant.records[0]
                mutant = np.where(g == 1)[0]
                sum_esizes[mutant] += pop.mutations[r.key].s
            ind = int(len(samples_at_time_u)/2)
            temp_gvalues = np.zeros(ind)
            temp_gvalues += sum_esizes[0::2]
            temp_gvalues += sum_esizes[1::2]
            genetic_values_from_ts.extend(temp_gvalues.tolist())
            genetic_trait_values_from_sim.extend(
                metadata['g'][np.where(metadata_record_times == u)[0]].tolist())

        for i, j in zip(genetic_trait_values_from_sim, genetic_values_from_ts):
            self.assertAlmostEqual(i, j)

        # This is a second test that we correctly stored metadata
        # in the correct order
        for i, j in zip(pop.ancient_sample_metadata, recorder.data):
            self.assertEqual(i.g, j[1].g)
            self.assertEqual(i.e, j[1].e)
            self.assertEqual(i.w, j[1].w)
            self.assertEqual(i.label, j[1].label)


class TestDataMatrixIterator(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # TODO add neutral variants
        self.N = 1000
        self.demography = np.array([self.N]*self.N, dtype=np.uint32)
        self.rho = 1.
        self.theta = 100.
        self.nreps = 500
        self.mu = self.theta/(4*self.N)
        self.r = self.rho/(4*self.N)

        self.GSS = fwdpy11.GSS(VS=1, opt=0)
        a = fwdpy11.Additive(2.0, self.GSS)
        self.p = {'nregions': [],
                  'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
                  'recregions': [fwdpy11.Region(0, 1, 1)],
                  'rates': (0.0, 0.025, self.r),
                  'gvalue': a,
                  'prune_selected': False,
                  'demography': self.demography
                  }
        self.params = fwdpy11.ModelParams(**self.p)
        self.rng = fwdpy11.GSLrng(101*45*110*210)
        self.pop = fwdpy11.DiploidPopulation(self.N, 1.0)
        self.all_samples = [i for i in range(2*self.N)]
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        self.dm = fwdpy11.data_matrix_from_tables(self.pop.tables,
                                                  self.all_samples,
                                                  True, True)
        self.neutral = np.array(self.dm.neutral)
        self.npos = np.array(self.dm.neutral.positions)
        self.selected = np.array(self.dm.selected)
        self.spos = np.array(self.dm.selected.positions)

    def test_entire_matrix(self):
        dmi = fwdpy11.DataMatrixIterator(self.pop.tables,
                                         self.all_samples,
                                         [(0, 1)], True, True)
        niterations = 0
        for dm in dmi:
            niterations += 1
            for i in dm.selected_keys:
                self.assertFalse(self.pop.mutations[i].neutral)
            self.assertTrue(np.array_equal(
                self.dm.selected_keys, dm.selected_keys))
            self.assertTrue(np.array_equal(dm.neutral, self.neutral))
            self.assertTrue(np.array_equal(dm.selected, self.selected))
        self.assertEqual(niterations, 1)

    def test_single_slice(self):
        dm = fwdpy11.DataMatrixIterator(self.pop.tables,
                                        self.all_samples,
                                        [(0.1, 0.2)], True, True)
        dm = next(dm)
        rows = np.where((self.spos >= 0.1) & (self.spos < 0.2))[0]
        pos_slice = self.spos[rows]
        selected_slice = self.selected[rows, ]
        self.assertTrue(np.array_equal(dm.selected_positions, pos_slice))
        self.assertTrue(np.array_equal(dm.selected, selected_slice))

    def test_nonoverlapping_slices(self):
        slices = [(0.1, 0.2), (0.21, 0.37), (0.5, 0.55)]
        dmi = fwdpy11.DataMatrixIterator(self.pop.tables,
                                         self.all_samples,
                                         slices, True, True)
        niterations = 0
        for r, dm in zip(slices, dmi):
            niterations += 1
            rows = np.where((self.spos >= r[0]) & (self.spos < r[1]))[0]
            pos_slice = self.spos[rows]
            selected_slice = self.selected[rows, ]
            self.assertTrue(np.array_equal(dm.selected_positions, pos_slice))
            self.assertTrue(np.array_equal(dm.selected, selected_slice))
        self.assertEqual(niterations, len(slices))

    def test_complex_slices(self):
        slices = [(0.1, 0.2), (0.15, 0.23), (0.21, 0.37),
                  (0.38, 0.5337), (0.5, 0.55)]
        dmi = fwdpy11.DataMatrixIterator(self.pop.tables,
                                         self.all_samples,
                                         slices, True, True)
        niterations = 0
        for r, dm in zip(slices, dmi):
            niterations += 1
            rows = np.where((self.spos >= r[0]) & (self.spos < r[1]))[0]
            pos_slice = self.spos[rows]
            selected_slice = self.selected[rows, ]
            self.assertTrue(np.array_equal(dm.selected_positions, pos_slice))
            self.assertTrue(np.array_equal(dm.selected, selected_slice))
        self.assertEqual(niterations, len(slices))

    def test_nested_slices(self):
        slices = [(0.1, 0.2), (0.15, 0.19), (0.21, 0.37),
                  (0.38, 0.5337), (0.39, 0.432), (0.5, 0.55)]
        dmi = fwdpy11.DataMatrixIterator(self.pop.tables,
                                         self.all_samples,
                                         slices, True, True)
        for r, dm in zip(slices, dmi):
            rows = np.where((self.spos >= r[0]) & (self.spos < r[1]))[0]
            pos_slice = self.spos[rows]
            selected_slice = self.selected[rows, ]
            self.assertTrue(np.array_equal(dm.selected_positions, pos_slice))
            self.assertTrue(np.array_equal(dm.selected, selected_slice))


class TestTreeSequenceResettingDuringTimeSeriesAnalysis(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        class CountSamplesPerTimePoint(object):
            def __init__(self):
                self.sample_timepoints = []
                self.sample_sizes = []
                self.timepoint_seen = {}

            def __call__(self, pop):
                assert len(pop.tables.preserved_nodes)//2 ==  \
                    len(pop.ancient_sample_metadata)
                # Get the most recent ancient samples
                # and record their number.  We do this
                # by a "brute-force" approach
                for t, n, m in pop.sample_timepoints(False):
                    if t not in self.timepoint_seen:
                        self.timepoint_seen[t] = 1
                    else:
                        self.timepoint_seen[t] += 1
                    if t not in self.sample_timepoints:
                        self.sample_timepoints.append(t)
                        self.sample_sizes.append(len(n)//2)

                    # simplify to each time point
                    tables, idmap = fwdpy11.simplify_tables(pop.tables, n)
                    for ni in n:
                        assert idmap[ni] != fwdpy11.NULL_NODE
                        assert tables.nodes[idmap[ni]].time == t

        self.N = 1000
        self.demography = np.array([self.N]*101, dtype=np.uint32)
        self.rho = 1.
        self.r = self.rho/(4*self.N)

        self.GSS = fwdpy11.GSS(VS=1, opt=0)
        a = fwdpy11.Additive(2.0, self.GSS)
        self.p = {'nregions': [],
                  'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
                  'recregions': [fwdpy11.Region(0, 1, 1)],
                  'rates': (0.0, 0.025, self.r),
                  'gvalue': a,
                  'prune_selected': False,
                  'demography': self.demography
                  }
        self.params = fwdpy11.ModelParams(**self.p)
        self.rng = fwdpy11.GSLrng(101*45*110*210)
        self.pop = fwdpy11.DiploidPopulation(self.N, 1.0)
        self.all_samples = [i for i in range(2*self.N)]
        self.ancient_sample_recorder = \
            fwdpy11.RandomAncientSamples(seed=42,
                                         samplesize=10,
                                         timepoints=[i for i in range(1, 101)])
        self.resetter = CountSamplesPerTimePoint()
        fwdpy11.evolvets(self.rng, self.pop, self.params, 5,
                         recorder=self.ancient_sample_recorder,
                         post_simplification_recorder=self.resetter)

    def test_no_preserved_nodes(self):
        self.assertEqual(len(self.pop.tables.preserved_nodes), 0)

    def test_no_ancient_sample_metadata(self):
        self.assertEqual(len(self.pop.ancient_sample_metadata), 0)

    def test_all_timepoints_seen_exactly_once(self):
        for i, j in self.resetter.timepoint_seen.items():
            self.assertEqual(j, 1)

    def test_all_expected_timepoints_are_present(self):
        self.assertEqual(self.resetter.sample_timepoints,
                         [i for i in range(1, 101)])

    def test_all_sample_sizes_are_correct(self):
        self.assertTrue(
            all([i == 10 for i in self.resetter.sample_sizes]) is True)


class TestDiploidPopulationInitialization(unittest.TestCase):
    def test_initial_nodes(self):
        pop = fwdpy11.DiploidPopulation(100, 1.)
        md = np.array(pop.diploid_metadata, copy=False)
        n = md['nodes'].flatten()
        self.assertTrue(np.array_equal(n, np.arange(2*pop.N, dtype=n.dtype)))


class TestTrackMutationCounts(unittest.TestCase):
    """
    Test for GitHub issue 326.

    Simplifying every generation w/strong positive selection is
    key to triggering the bug.
    """
    @classmethod
    def setUp(self):
        class FreqTracker(object):
            def __init__(self):
                self.freqs = dict()

            def __call__(self, pop, recorder):
                for i, j in enumerate(pop.mcounts):
                    if j == 2*pop.N:
                        k = pop.mutations[i].key
                        if k in self.freqs:
                            if self.freqs[k][-1][1] < 2*pop.N:
                                self.freqs[k].append((pop.generation, j))
                        else:
                            self.freqs[k] = [(pop.generation, j)]

        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        self.ft = FreqTracker()

    def test_without_table_indexing(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params,
                         1, self.ft, track_mutation_counts=True,
                         suppress_table_indexing=True)
        assert len(
            self.pop.fixations) > 0, \
            "Nothing fixed, so test case is not helpful"

        for i, j in zip(self.pop.fixations, self.pop.fixation_times):
            self.assertTrue(i.key in self.ft.freqs)
            self.assertEqual(self.ft.freqs[i.key][-1][0], j)

    def test_with_table_indexing(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params,
                         1, self.ft, track_mutation_counts=True)
        assert len(
            self.pop.fixations) > 0, \
            "Nothing fixed, so test case is not helpful"

        for i, j in zip(self.pop.fixations, self.pop.fixation_times):
            self.assertTrue(i.key in self.ft.freqs)
            self.assertEqual(self.ft.freqs[i.key][-1][0], j)


if __name__ == "__main__":
    unittest.main()



import fwdpy11
import numpy as np
import inherit_noise
import unittest


class TestInheritNoise(unittest.TestCase):
    """
    The class inherit_noise.TestInheritNoise passes the
    sum of all generations into the "e" field of
    offspring metadata.  Thus, if offspring metadata
    can see parental metadata, then we can test that the
    final alive individuals (in a sim of non-overlapping
    generations) all have e fields that are the sum of
    the generations that have occurred.
    """
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': (0, 0, 0),
                      'demography': np.array([self.pop.N]*3, dtype=np.uint32),
                      'gvalue': fwdpy11.Additive(2., fwdpy11.GSS(0, 1),
                                                 inherit_noise.IneritedNoise())
                      }

    def test_noise_values_tree_sequences(self):
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertEqual(self.pop.generation, 3)
        self.assertTrue(
            all([i.e == 6 for i in self.pop.diploid_metadata]) is True)

    def test_noise_values_without_tree_sequences(self):
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolve_genomes(self.rng, self.pop, params)
        self.assertEqual(self.pop.generation, 3)
        self.assertTrue(
            all([i.e == 6 for i in self.pop.diploid_metadata]) is True)


if __name__ == "__main__":
    unittest.main()



import fwdpy11 as fp11
import fwdpy11
import fwdpy11.ezparams as fp11ez
# concurrent.futures is Python 3 only
import concurrent.futures as cf
import numpy as np
import unittest


def evolve_and_return(args):
    """
    This function runs our simulation.
    The input arguments come in a tuple,
    which is required by many of Python's
    functions for execution in separate processes.

    For this function, the arguments are the population
    size and a random number seed.
    """
    from fwdpy11 import Multiplicative
    N, seed = args
    # Construct as single-deme object
    # with N diploids
    pop = fp11.DiploidPopulation(N)
    theta = 100.0
    # Initialize a random number generator
    rng = fp11.GSLrng(seed)
    p = fp11ez.mslike(pop, simlen=100, rates=(
        theta / float(4 * pop.N), 1e-3, theta / float(4 * pop.N)))
    p['gvalue'] = Multiplicative(2.)
    params = fp11.ModelParams(**p)
    fp11.evolve_genomes(rng, pop, params)
    # The population is picklable, and so
    # we can return it from another process
    return pop


class testSimpleParallel(unittest.TestCase):
    """
    This would be your __name__=="__main__"
    in a script
    """

    def test_simple_parallel(self):
        np.random.seed(101)
        # Generate a list of arguments for our processes.
        # We generation 10 random number seeds
        args = [(1000, seed) for seed in np.random.randint(0, 42000000, 10)]
        # Use a process pool with a max of 10 workers
        with cf.ProcessPoolExecutor(10) as pool:
            # Run our simulations and get the
            # result back, which will be
            # the population
            for res in pool.map(evolve_and_return, args):
                self.assertEqual(res.N, 1000)
                self.assertEqual(res.generation, 100)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest

# TODO reimplement to test new API

# import fwdpy11 as fp11
# import fwdpy11 as fp11mp
# import fwdpy11.multilocus as ml
# import fwdpy11.wright_fisher_qtrait as wfq
# from fwdpy11.fitness import DiploidAdditive
# from quick_pops import quick_slocus_qtrait_pop_params
# import numpy as np
# import pickle
# 
# 
# class testDiploidParams(unittest.TestCase):
#     @classmethod
#     def setUpClass(self):
#         # We do not assign certain attributes,
#         # which allows us to test the defaults.
#         self.pdict = {'nregions': [],
#                       'sregions': [],
#                       'recregions': [],
#                       'rates': [0, 0, 0],
#                       'demography': np.array([100] * 100, dtype=np.uint32)
#                       }
# 
#     def test_DiploidParamsDefaults(self):
#         from fwdpy11.fitness import DiploidMult
#         m = fp11mp.DiploidParams(**self.pdict)
#         self.assertEqual(type(m.gvalue), DiploidMult)
# 
#     def test_DiploidParamsQDefaults(self):
#         self.pdict['prune_selected'] = False  # Suppresses a warning
#         from fwdpy11.trait_values import DiploidAdditiveTrait
#         from fwdpy11.wright_fisher_qtrait import GSS
#         m = fp11mp.DiploidParamsQ(**self.pdict)
#         self.assertEqual(type(m.gvalue), DiploidAdditiveTrait)
#         self.assertEqual(type(m.trait2w), GSS)
#         self.assertEqual(m.trait2w.VS, 1.0)
#         self.assertEqual(m.trait2w.O, 0.0)
# 
#     def test_set_nregions(self):
#         self.setUpClass()
#         m = fp11mp.DiploidParams(**self.pdict)
#         try:
#             m.nregions = [fp11.Region(0, 1, 1)]
#         except:
#             self.fail("unexpected exception")
#         # An empty list is ok
#         try:
#             m.nregions = []
#         except:
#             self.fail("unexpected exception")
#         # TypeError b/c not iterable:
#         with self.assertRaises(TypeError):
#             m.nregions = fp11.Region(0, 1, 1)
#         with self.assertRaises(TypeError):
#             m.nregions = [fp11.ExpS(0, 1, 1, 0.25)]
# 
#     def test_set_recregions(self):
#         self.setUpClass()
#         m = fp11mp.DiploidParams(**self.pdict)
#         try:
#             m.recregions = [fp11.Region(0, 1, 1)]
#         except:
#             self.fail("unexpected exception")
#         # An empty list is ok
#         try:
#             m.recregions = []
#         except:
#             self.fail("unexpected exception")
#         # TypeError b/c not iterable:
#         with self.assertRaises(TypeError):
#             m.recregions = fp11.Region(0, 1, 1)
#         with self.assertRaises(TypeError):
#             m.recregions = [fp11.ExpS(0, 1, 1, 0.25)]
# 
#     def test_set_sregions(self):
#         self.setUpClass()
#         m = fp11mp.DiploidParams(**self.pdict)
#         try:
#             m.sregions = [fp11.GaussianS(0, 1, 1, 0.1)]
#         except:
#             self.fail("unexpected exception")
#         # An empty list is ok
#         try:
#             m.sregions = []
#         except:
#             self.fail("unexpected exception")
#         with self.assertRaises(TypeError):
#             m.sregions = fp11.GaussianS(0, 1, 1, 0.2)
#         with self.assertRaises(TypeError):
#             m.sregions = [fp11.Region(0, 1, 1)]
# 
#     def test_init_demog(self):
#         self.setUpClass()
#         try:
#             m = fp11mp.DiploidParams(**self.pdict)
#             m
#         except:
#             self.fail("unexpected exception")
# 
#     def test_init_demog_list(self):
#         self.pdict['demography'] = [100] * 10
#         with self.assertRaises(ValueError):
#             m = fp11mp.DiploidParams(**self.pdict)
#             m
# 
#     def test_init_bad_popsizes(self):
#         self.pdict['demography'] = np.array([100] * 10 + [-1])
#         with self.assertRaises(ValueError):
#             m = fp11mp.DiploidParams(**self.pdict)
#             m
# 
#     def test_bad_rates(self):
#         self.setUpClass()
#         self.pdict['mutrate_n'] = None
#         with self.assertRaises(ValueError):
#             m = fp11mp.DiploidParams(**self.pdict)
#             m
#         self.pdict['mutrate_n'] = -0.01
#         with self.assertRaises(ValueError):
#             m = fp11mp.DiploidParams(**self.pdict)
#             m
# 
#     def test_warning_for_zero_rate(self):
#         # We pass in some sregions but leave mutrate_s == 0.
#         self.setUpClass()
#         self.pdict['sregions'] = [fp11.GammaS(0, 1, 1, -0.1, 0.10, 1.0)]
#         with self.assertWarns(UserWarning):
#             m = fp11mp.DiploidParams(**self.pdict)
#             m.validate()
# 
#     def test_set_rates_list(self):
#         self.setUpClass()
#         self.pdict['rates'] = [0., 1e-3, 0.25]
#         try:
#             m = fp11mp.DiploidParams(**self.pdict)
#             m
#         except:
#             self.fail("unexpected exception")
# 
#     def test_set_rates_tuple(self):
#         m = fp11mp.DiploidParams()
#         m.rates = (0., 1e-3, 0.25)
#         self.assertEqual(m.mutrate_n, 0.0)
#         self.assertEqual(m.mutrate_s, 1e-3)
#         self.assertEqual(m.recrate, 0.25)
#         with self.assertRaises(ValueError):
#             m.rates = (0., -1e-3, 0.25)
# 
#     def test_set_rates_dict(self):
#         m = fp11mp.DiploidParams()
#         with self.assertRaises(ValueError):
#             m.rates = {'mutrate_n': 1e-3}
#         with self.assertRaises(ValueError):
#             m.rates = {'mutrate_n': -1e-3,
#                        'mutrate_s': 1e-3,
#                        'recrate': 1e-3}
#         with self.assertRaises(ValueError):
#             m.rates = {'monkeys_n': 1e-3}
# 
#     def test_set_gvalue(self):
#         m = fp11mp.DiploidParams()
#         from fwdpy11.fitness import DiploidAdditive
#         m.gvalue = DiploidAdditive()
# 
#     def test_set_bad_gvalue(self):
#         m = fp11mp.DiploidParams()
#         with self.assertRaises(ValueError):
#             m.gvalue = []
# 
#     def test_invalid_kwargs(self):
#         from fwdpy11.fitness import DiploidAdditive
#         with self.assertRaises(ValueError):
#             m = fp11mp.DiploidParams(fitness=DiploidAdditive())
#             m
#
# if __name__ == "__main__":
#     pass



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

"""
This is basically a test of bad models.
"""

import fwdpy11
import unittest
import numpy as np


def setup_and_run_model(pop, ddemog, simlen, recorder=None, seed=654321):
    pdict = {'nregions': [],
             'sregions': [],
             'recregions': [],
             'rates': (0, 0, 0,),
             'gvalue': fwdpy11.Multiplicative(2.),
             'demography': ddemog,
             'simlen': simlen
             }
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(seed)
    fwdpy11.evolvets(rng, pop, params, 100, recorder)


class TestBadMassMigrations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.)

    def test_move_too_many(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.move_individuals(1, 0, 1, 1),
             fwdpy11.move_individuals(2, 0, 1, 1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)
        with self.assertRaises(RuntimeError):
            setup_and_run_model(self.pop, d, 5)

    def test_copy_from_empty_deme(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1),
             fwdpy11.copy_individuals(1, 0, 1, 1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)
        with self.assertRaises(RuntimeError):
            setup_and_run_model(self.pop, d, 5)


class TestDetectingExtinctions(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.)

    def test_no_valid_parents_V1(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_V2(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(1, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_V3(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 0),
             fwdpy11.SetDemeSize(11, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_with_migration(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       migmatrix=M)
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)

    def test_no_valid_parents_with_migration_V2(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        cM = [
            fwdpy11.SetMigrationRates(0, np.array([0, 1, 0, 1]).reshape(2, 2))
        ]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       set_migration_rates=cM,
                                       migmatrix=M)
        try:
            fwdpy11.DemographyDebugger(self.pop, d)
        except Exception:
            self.fail("unexpected exception")
        setup_and_run_model(self.pop, d, 5)

    def test_no_valid_parents_with_migration_V2_migration_rescue(self):
        """
        This is a convoluted model:
        0 moves 100% to 1. Thus, there are no parents anymore for 0.
        But, this is okay because the new deme 1 will provide parents
        back into 0, thus "un-extincting" 0.  Then, in the next
        generation, each deme gets 50/50 of its ancestry from each
        deme.
        """
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        cM = [
            fwdpy11.SetMigrationRates(0, np.array([0, 1, 0, 1]).reshape(2, 2)),
            fwdpy11.SetMigrationRates(1, np.array([0.5]*4).reshape(2, 2)),
        ]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       set_migration_rates=cM,
                                       migmatrix=M)
        try:
            fwdpy11.DemographyDebugger(self.pop, d)
        except Exception:
            self.fail("unexpected exception")
        setup_and_run_model(self.pop, d, 5)

    def test_no_valid_parents_with_migration_V3(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       migmatrix=(M, True))
        with self.assertRaises(ValueError):
            fwdpy11.DemographyDebugger(self.pop, d)
        with self.assertRaises(fwdpy11.DemographyError):
            setup_and_run_model(self.pop, d, 5)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2019 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Local adaptation of a quantitative trait to differing optima.
"""

import fwdpy11
import pandas as pd
import math
import numpy as np
import sys
import argparse
from collections import namedtuple

# Simulations with tree sequence recording need
# to know the max position in a genome.  Here,
# we use a length of 1.0. Thus, all mutation
# and recombination events will be uniform
# random variables on the continuous interval
# [0, GENOME_LENGTH).
GENOME_LENGTH = 1.0

# When recording quant-genetic statistics during a simulation,
# we will use this type. Named tuples are extremely efficient,
# and they are easily converted into Pandas DataFrame objects,
# which is very convenient for analysis and output.
Datum = namedtuple('Data', ['generation', 'deme', 'gbar', 'vg', 'wbar'])


def make_parser():
    """
    Create a command-line interface to the script.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    required = parser.add_argument_group("Required arguments")
    required.add_argument("--popsize", "-N", type=int,
                          help="Diploid population size")
    required.add_argument("--mu", "-m", type=float,
                          help="Mutation rate (per gamete, per generation)")
    required.add_argument("--sigma", "-s", type=float,
                          help="Standard deviation of Gaussian"
                          "distribution of mutational effects")

    optional = parser.add_argument_group("Optional arguments")
    optional.add_argument("--rho", type=float, default=1000.0,
                          help="Scaled recombination rate, rho=4Nr")
    optional.add_argument("--VS", type=float, default=10.0,
                          help="Inverse strength of stabilizing selection")
    optional.add_argument("--opt", type=float, default=1.0,
                          help="Value of new phenotypic optimum")
    optional.add_argument("--migrates", type=float, nargs=2, default=None,
                          help="Migration rates from 0 to 1 and 1 to 0, respectively.")
    optional.add_argument("--time", type=float, default=0.1,
                          help="Amount of time to simulate past"
                          "optimum shift, in units of N")
    optional.add_argument("--plotfile", type=str, default=None,
                          help="File name for plot")
    optional.add_argument("--seed", type=int, default=42,
                          help="Random number seed.")

    return parser


def validate_arguments(args):
    """
    Validate input arguments.
    Note: this is likely incomplete.
    """
    if args.popsize is None:
        raise ValueError("popsize cannot be None")
    if args.mu < 0:
        raise ValueError("mu must be non-negative")
    if args.mu is None:
        raise ValueError("mu cannot be None")
    if args.mu < 0 or math.isfinite(args.mu) is False:
        raise ValueError("Mutation rate must be non-negative and finite")
    if args.sigma is None:
        raise ValueError("sigma cannot be none")
    if args.sigma < 0 or math.isfinite(args.sigma) is False:
        raise ValueError("Std. dev. of distribution of effect sizes"
                         "must be non-negative and finite")

    if args.migrates is not None:
        for m in args.migrates:
            if m < 0 or m > 1:
                raise ValueError("migration rates must be 0 <= m <= 1")


def make_migmatrix(migrates):
    if migrates is None:
        return None
    mm = np.zeros(4).reshape(2, 2)
    mm[0, 1] = migrates[1]
    mm[1, 0] = migrates[0]
    rs = np.sum(mm, axis=1)
    np.fill_diagonal(mm, 1.0 - rs)
    return fwdpy11.MigrationMatrix(mm)


class Recorder(object):
    """
    fwdpy11 allows you to define objects that record data
    from populations during simulation.  Such objects must
    be callable, and the easiest way to do things is to
    create a class with a __call__ function.
    """

    def __init__(self, start):
        self.data = []
        self.start = start

    def __call__(self, pop, recorder):
        if pop.generation >= self.start:
            # Record mean trait value each generation.
            md = np.array(pop.diploid_metadata, copy=False)
            demes = np.unique(md['deme'])
            for d in demes:
                w = np.where(md['deme'] == d)[0]
                gbar = md['g'][w].mean()
                vg = md['g'][w].var()
                wbar = md['w'][w].mean()
                self.data.append(Datum(pop.generation, d,
                                       gbar, vg, wbar))


def plot_output(data, filename):
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
    fig = plt.figure(figsize=(9, 3))
    gs = gridspec.GridSpec(ncols=3, nrows=1, figure=fig)
    ax_gbar = fig.add_subplot(gs[0, 0])
    ax_vg = fig.add_subplot(gs[0, 1])
    ax_wbar = fig.add_subplot(gs[0, 2])

    df = pd.DataFrame(data, columns=Datum._fields)
    g = df.groupby(['deme'])
    for n, gi in g:
        ax_gbar.plot(gi['generation'], gi['gbar'], label="Deme {}".format(n))
        ax_vg.plot(gi['generation'], gi['vg'], label="Deme {}".format(n))
        ax_wbar.plot(gi['generation'], gi['wbar'], label="Deme {}".format(n))

    for ax in [ax_gbar, ax_vg, ax_wbar]:
        ax.set_xlabel("Generation")

    ax_gbar.set_ylabel(r'$\bar{g}$')
    ax_vg.set_ylabel(r'$V(G)$')
    ax_wbar.set_ylabel(r'$\bar{w}$')
    ax_gbar.legend()
    plt.tight_layout()
    plt.savefig(filename)


def runsim(args):
    """
    Run the simulation.
    """
    pop = fwdpy11.DiploidPopulation(2*args.popsize, GENOME_LENGTH)

    np.random.seed(args.seed)
    rng = fwdpy11.GSLrng(args.seed)

    GSSmo0 = fwdpy11.GSSmo(
        [(0, 0, args.VS), (10*args.popsize, args.opt, args.VS)])
    GSSmo1 = fwdpy11.GSSmo(
        [(0, 0, args.VS), (10*args.popsize, -1.0*args.opt, args.VS)])

    mm = make_migmatrix(args.migrates)
    dd = fwdpy11.DiscreteDemography(
        mass_migrations=[fwdpy11.move_individuals(0, 0, 1, 0.5)],
        migmatrix=mm)

    p = {'nregions': [],  # No neutral mutations -- add them later!
         'gvalue': [fwdpy11.Additive(2.0, GSSmo0),
                    fwdpy11.Additive(2.0, GSSmo1)],
         'sregions': [fwdpy11.GaussianS(0, GENOME_LENGTH, 1, args.sigma)],
         'recregions': [fwdpy11.Region(0, GENOME_LENGTH, 1)],
         'rates': (0.0, args.mu, args.rho/float(4*args.popsize)),
         # Keep mutations at frequency 1 in the pop if they affect fitness.
         'prune_selected': False,
         'demography': dd,
         'simlen': 10*args.popsize + int(args.popsize*args.time)
         }
    params = fwdpy11.ModelParams(**p)

    r = Recorder(10*args.popsize)
    fwdpy11.evolvets(rng, pop, params, 100, r, suppress_table_indexing=True)
    if args.plotfile is not None:
        plot_output(r.data, args.plotfile)


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args(sys.argv[1:])
    validate_arguments(args)
    runsim(args)



import unittest
import fwdpy11
import numpy as np
import copy
from test_tree_sequences import set_up_quant_trait_model
from test_tree_sequences import set_up_standard_pop_gen_model


def _count_mutations_from_diploids(pop):
    mc = np.zeros(len(pop.mutations), dtype=np.uint32)
    for d in pop.diploids:
        for k in pop.haploid_genomes[d.first].mutations:
            mc[k] += 1
        for k in pop.haploid_genomes[d.first].smutations:
            mc[k] += 1
        for k in pop.haploid_genomes[d.second].mutations:
            mc[k] += 1
        for k in pop.haploid_genomes[d.second].smutations:
            mc[k] += 1
    return mc


def _compare_counts_for_nonneutral_variants(pop, mc):
    w = np.array(
        [i for i, j in enumerate(pop.mutations) if j.neutral is False], dtype=np.uint32)
    return all(pop.mcounts[w] == mc[w]) is True


class TestKeepFixations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop = set_up_quant_trait_model()
        self.params.rates = (1e-3, self.params.rates[1], self.params.rates[2])
        self.params.nregions = [fwdpy11.Region(0, 1, 1)]

    def test_mutation_counts_with_indexing(self):
        """
        Tests atypical use case where neutral mutations are placed into genomes
        """
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))

    def test_mutation_counts_with_indexing_suppressed(self):
        """
        Tests atypical use case where neutral mutations are placed into genomes
        """
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100,
                         suppress_table_indexing=True)
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))

    def test_mutation_counts_with_indexing_suppressed_no_neutral_muts_in_genomes(self):
        """
        A sim w/ and w/o putting neutral variants in genomes
        should give the same mutation counts.
        """
        pop2 = copy.deepcopy(self.pop)
        rng = fwdpy11.GSLrng(101*45*110*210)  # Use same seed!!!
        self.params.prune_selected = False
        params = copy.deepcopy(self.params)
        fwdpy11.evolvets(rng, pop2, params, 100,
                         suppress_table_indexing=True)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100,
                         suppress_table_indexing=True)
        ti = fwdpy11.TreeIterator(
            self.pop.tables, [i for i in range(2*self.pop.N)])
        mc = _count_mutations_from_diploids(self.pop)
        for t in ti:
            for m in t.mutations():
                # Have to skip neutral mutations b/c they won't
                # end up in mc b/c it is obtained from genomes
                if pop2.mutations[m.key].neutral is False:
                    self.assertEqual(mc[m.key], self.pop.mcounts[m.key])
                self.assertEqual(t.leaf_counts(m.node), pop2.mcounts[m.key])


class TestPruneFixations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        self.params.rates = (1e-3, self.params.rates[1], self.params.rates[2])
        self.params.nregions = [fwdpy11.Region(0, 1, 1)]

    def test_no_mutations(self):
        self.params.rates = (0, 0, 0)
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)

    def test_mutation_counts_with_indexing(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))
        self.assertEqual(len(np.where(mc == 2*self.pop.N)[0]), 0)

    def test_mutation_counts_with_indexing_suppressed(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100,
                         suppress_table_indexing=True)
        mc = _count_mutations_from_diploids(self.pop)
        self.assertTrue(_compare_counts_for_nonneutral_variants(self.pop, mc))
        self.assertEqual(len(np.where(mc == 2*self.pop.N)[0]), 0)


class TestNeutralMutRegions(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.params, self.rng, self.pop = set_up_standard_pop_gen_model()
        self.params.rates = (1e-3, self.params.rates[1], self.params.rates[2])
        self.params.nregions = [fwdpy11.Region(0, 0.25, 1),
                                fwdpy11.Region(0.5, 1, 1)]

    def test_neutral_mut_locations(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 100)
        pos = [self.pop.tables.sites[i.site].position for
               i in self.pop.tables.mutations if i.neutral]
        self.assertTrue(all([i < 0.25 or i >= 0.5 for i in pos]))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2019 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import fwdpy11
import unittest
import numpy as np


def setup_pdict(demog, simlen):
    nregions = []
    sregions = []
    recregions = []

    pdict = {'nregions': nregions,
             'sregions': sregions,
             'recregions': recregions,
             'rates': (0, 0, None),
             'gvalue': fwdpy11.Multiplicative(2.),
             'demography': demog,
             'simlen': simlen,
             'prune_selected': True
             }
    return pdict


def setup_pop_rng(N=100, seed=42):
    pop = fwdpy11.DiploidPopulation(N, 1.0)
    rng = fwdpy11.GSLrng(seed)
    return pop, rng


class TestTwoDemeIMModel(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.Nanc = 100
        self.N0, self.N1 = 1.1, 3.0
        from fwdpy11.demographic_models.IM import two_deme_IM
        self.d, self.t1, self.t2 = two_deme_IM(self.Nanc, 0.1, 0.7,
                                               (self.N0, self.N1),
                                               (1e-2, 0.25),
                                               burnin=1.0)
        self.pop, self.rng = setup_pop_rng(self.Nanc)

    def test_complete_sim(self):
        pdict = setup_pdict(self.d, self.t1+self.t2)
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, params.simlen)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], np.rint(self.N0*self.Nanc).astype(int))
        self.assertEqual(deme_sizes[1], np.rint(self.N1*self.Nanc).astype(int))

    def test_evolve_in_two_steps(self):
        pdict = setup_pdict(self.d, self.t1)
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, self.t1)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], self.Nanc)
        self.assertTrue(1 not in deme_sizes)

        params.simlen = self.t2
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, self.t1+self.t2)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], np.rint(self.N0*self.Nanc).astype(int))
        self.assertEqual(deme_sizes[1], np.rint(self.N1*self.Nanc).astype(int))

    def test_evolve_in_two_steps_restart_with_two_demes(self):
        deltat = 2
        pdict = setup_pdict(self.d, self.t1+deltat)
        params = fwdpy11.ModelParams(**pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, self.t1+deltat)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertTrue(0 in deme_sizes)
        self.assertTrue(1 in deme_sizes)

        params.simlen = self.t2-deltat
        fwdpy11.evolvets(self.rng, self.pop, params, 10)
        self.assertEqual(self.pop.generation, self.t1+self.t2)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes[0], np.rint(self.N0*self.Nanc).astype(int))
        self.assertEqual(deme_sizes[1], np.rint(self.N1*self.Nanc).astype(int))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


def evolvets(rng, pop, params, simplification_interval, recorder=None,
             post_simplification_recorder=None,
             suppress_table_indexing=False, record_gvalue_matrix=False,
             stopping_criterion=None,
             track_mutation_counts=False,
             remove_extinct_variants=True):
    """
    Evolve a population with tree sequence recording

    :param rng: random number generator
    :type rng: :class:`fwdpy11.GSLrng`
    :param pop: A population
    :type pop: :class:`fwdpy11.DiploidPopulation`
    :param params: simulation parameters
    :type params: :class:`fwdpy11.ModelParams`
    :param simplification_interval: Number of generations between simplifications.
    :type simplification_interval: int
    :param recorder: (None) A temporal sampler/data recorder.
    :type recorder: callable
    :param post_simplification_recorder: (None) A temporal sampler
    :type post_simplification_recorder: callable
    :param suppress_table_indexing: (False) Prevents edge table indexing until end of simulation
    :type suppress_table_indexing: boolean
    :param record_gvalue_matrix: (False) Whether to record genetic values into :attr:`fwdpy11.PopulationBase.genetic_values`.
    :type record_gvalue_matrix: boolean

    The recording of genetic values into :attr:`fwdpy11.PopulationBase.genetic_values` is suppressed by default.  First, it
    is redundant with :attr:`fwdpy11.DiploidMetadata.g` for the common case of mutational effects on a single trait.
    Second, we save some memory by not tracking these matrices.  However, it is useful to track these data for some
    cases when simulating multivariate mutational effects (pleiotropy).

    For a detailed description of `post_simplification_recorder`, see :ref:`tstimeseries`.

    .. note::
        If recorder is None,
        then :class:`fwdpy11.NoAncientSamples` will be used.

        If post_simplification_recorder is None, then
        :class:`fwdpy11.RecordNothing` will be used.

    .. versionchanged:: 0.5.2

        Added post_simplification_recorder.

    """
    import warnings

    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Will throw exception if anything is wrong:
        params.validate()

    if recorder is None:
        from ._fwdpy11 import NoAncientSamples
        recorder = NoAncientSamples()

    if post_simplification_recorder is None:
        from ._fwdpy11 import RecordNothing
        post_simplification_recorder = RecordNothing()
        reset_treeseqs_after_simplify = False
    else:
        reset_treeseqs_after_simplify = True

    if stopping_criterion is None:
        from ._fwdpy11 import _no_stopping
        stopping_criterion = _no_stopping

    from ._fwdpy11 import MutationRegions
    from ._fwdpy11 import dispatch_create_GeneticMap
    from ._fwdpy11 import evolve_with_tree_sequences
    pneutral = 0.0
    if params.mutrate_n + params.mutrate_s > 0.:
        pneutral = params.mutrate_n/(params.mutrate_n + params.mutrate_s)
    mm = MutationRegions.create(pneutral, params.nregions, params.sregions)
    rm = dispatch_create_GeneticMap(params.recrate, params.recregions)

    from ._fwdpy11 import SampleRecorder
    sr = SampleRecorder()
    from ._fwdpy11 import _dgvalue_pointer_vector
    gvpointers = _dgvalue_pointer_vector(params.gvalue)
    evolve_with_tree_sequences(rng, pop, sr, simplification_interval,
                               params.demography, params.simlen,
                               params.mutrate_n, params.mutrate_s,
                               mm, rm, gvpointers,
                               recorder, stopping_criterion,
                               params.pself, params.prune_selected is False,
                               suppress_table_indexing, record_gvalue_matrix,
                               track_mutation_counts,
                               remove_extinct_variants,
                               reset_treeseqs_after_simplify,
                               post_simplification_recorder)



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import fwdpy11
import warnings
import collections
import numpy as np


class DemographyDebugger(object):
    """
    Efficiently debug demographic events.

    The class rapidly executes a mock simulation
    of demographic events in order to try to rapidly
    detect errors. It also generates a "report"
    of the model details for double-checking.

    If model errors are found, a ``ValueError`` exception
    is raised. The types of errors are the same as would
    result in an exception during a simulation.

    .. versionadded:: 0.6.0
    """

    def __init__(self, pop, events, simlen=None, deme_labels=None):
        """
        :param pop: A population
        :type pop: :class:`fwdpy11.DiploidPopulation`
        :param events: The demographic events
        :type events: :class:`fwdpy11.DiscreteDemography`
        :param simlen: The total length of the simulation
        :type simlen: int
        :deme labels: A map from deme index to some other
                      kind of printable name
        :type labels: dict

        If ``simlen`` is provided, and it is greater than the time
        of the last event, then the final deme sizes will be calculated
        and reported.
        """
        # The setup
        self.M = None
        if events.migmatrix is not None:
            self.M = np.copy(events.migmatrix.M)
        self._validate_migration_rate_change_lengths(events)

        self.maxdemes = self._get_maxdemes(pop, events)

        if self.maxdemes < 1:
            raise ValueError("Invalid number of "
                             "demes in simulation: {}".format(self.maxdemes))

        self.current_deme_sizes = np.zeros(self.maxdemes, dtype=np.uint32)
        for i, j in zip(*pop.deme_sizes()):
            self.current_deme_sizes[i] = j

        self.growth_rates = np.array([fwdpy11.NOGROWTH]*self.maxdemes)
        self.growth_onset_times = np.zeros(self.maxdemes, dtype=np.uint32)
        self.selfing_rates = np.zeros(self.maxdemes)
        self.growth_initial_sizes = np.copy(self.current_deme_sizes)
        # NOTE/FIXME: some of the logic re: went_extinct and has_metadata
        # is redundant, convoluted, or both.
        self.went_extinct = np.zeros(self.maxdemes, dtype=np.int32)
        self.has_metadata = np.zeros(self.maxdemes, dtype=np.int32)
        self.has_metadata[(self.current_deme_sizes > 0)] = 1
        self.deme_labels = deme_labels

        # The real work
        self._report = None
        self._process_demographic_model(events, simlen)

    def _get_maxdemes(self, pop, events):
        """
        The maximum number of demes the sim can ever see.
        """
        md = np.array(pop.diploid_metadata, copy=False)
        max_from_md = md['deme'].max()
        max_from_events = -1

        def update_max_from_events(m, e):
            for i in e:
                try:
                    m = max(m, i.deme)
                except AttributeError:
                    m = max(m, i.source)
                    m = max(m, i.destination)
            return m

        for i in [events.mass_migrations, events.set_growth_rates,
                  events.set_deme_sizes, events.set_selfing_rates,
                  events.set_migration_rates]:
            max_from_events = update_max_from_events(max_from_events, i)

        current_max = max(max_from_md, max_from_events) + 1
        if self.M is None:
            return current_max

        if self.M.shape[0] < current_max:
            raise ValueError("The MigrationMatrix shape, {}, "
                             "does not match the max number of "
                             "demes present in the "
                             "simulation, ".format(self.M.shape, current_max))

        return max(current_max, self.M.shape[0])

    def _validate_migration_rate_change_lengths(self, events):
        """
        Various checks on the lengths of new migration rates.
        """
        if self.M is None and len(events.set_migration_rates) > 0:
            raise ValueError("migration rate changes are "
                             "set but there is no migration matrix")
        for i in events.set_migration_rates:
            if i.deme >= 0:
                if len(i.migrates) != self.M.shape[0]:
                    raise ValueError("Migration rates mismatch")
            else:  # Are replacing the entire matrix
                if len(i.migrates.flatten()) != len(self.M.flatten()):
                    raise ValueError("Migration rates mismatch")

    def _get_event_names(self, events):
        # NOTE: we rely on the fact that all public fields of
        # DiscreteDemography are event lists or the migration
        # matrix.  Thus, we filter out migmatrix and all
        # magic fxns.
        return [i for i in events.__dir__() if '__' not in i
                and 'migmatrix' not in i]

    def _make_event_queues(self, events):
        """
        Take references from the input so that we can process them
        w/o affecting the input.
        """
        from collections import deque

        rv = {}
        for e in self._get_event_names(events):
            rv[e] = deque([i for i in events.__getattribute__(e)])

        return rv

    def _get_next_event_time(self, event_queues):
        t = None
        for key, value in event_queues.items():
            if len(value) > 0:
                if t is None:
                    t = value[0].when
                else:
                    t = min(t, value[0].when)
        return t

    def _current_events(self, t, event_queues, event):
        elist = []
        while len(event_queues[event]) > 0 and \
                event_queues[event][0].when == t:
            e = event_queues[event][0]
            elist.append(e)
            event_queues[event].popleft()
        return elist

    def _label_deme(self, d):
        if self.deme_labels is None:
            return d
        if d not in self.deme_labels:
            return d
        return self.deme_labels[d]

    def _format_deme_sizes(self, sizes):
        rv = collections.OrderedDict()
        for i, j in enumerate(sizes):
            rv[self._label_deme(i)] = j
        return [(i, j) for i, j in rv.items()]

    def _apply_MassMigration(self, t, event_queues):
        for e in self._current_events(t, event_queues, 'mass_migrations'):
            if self.current_deme_sizes[e.source] == 0:
                temp = (e, self._label_deme(e.source))
                raise ValueError(
                    "mass migration at time {} involves "
                    "empty source deme {}".format(*temp))
            # NOTE: what is the C++ back-end doing if n_from_source < 1?
            # Answer: currently, that passes silently--is that okay?
            n_from_source = np.rint(
                self.current_deme_sizes[e.source]*e.fraction).astype(int)
            if n_from_source == 0:
                # Assume that this occurs from rounding issues,
                # so we treat it as a warning rather than an error
                temp = (t, self._label_deme(e.source),
                        self._label_deme(e.destination))
                message = "mass migration at time {} from {} " + \
                          "to {} moves no individuals"
                warnings.warn(message.format(*temp))
            if n_from_source > self.current_deme_sizes[e.source] and \
                    e.move_individuals is True:
                temp = (t, n_from_source, self._label_deme(e.source))
                raise ValueError("mass migration at time {} "
                                 "moves {} individuals from "
                                 "source deme {}".format(*temp))
            if e.move_individuals is True:
                self.current_deme_sizes[e.source] -= n_from_source
                self.current_deme_sizes[e.destination] += n_from_source
                temp = (n_from_source,
                        self._label_deme(e.source),
                        self._label_deme(e.destination))
                self._report.append("\tMass movement of {} "
                                    "from {} to {}\n".format(*temp))
                if n_from_source > 0:
                    self.has_metadata[e.destination] = 1
                if e.fraction == 1.0:
                    self.went_extinct[e.source] = 1
                    self.has_metadata[e.source] = 0
            else:  # A copy
                if n_from_source > 0:
                    self.has_metadata[e.destination] = 1
                self.current_deme_sizes[e.destination] += n_from_source
                temp = (n_from_source,
                        self._label_deme(e.source),
                        self._label_deme(e.destination))
                self._report.append("\tMass copy of {} "
                                    "from {} to {}\n".format(*temp))
            if e.resets_growth_rate is True:
                self.growth_rates[e.source] = fwdpy11.NOGROWTH
                self.growth_rates[e.destination] = fwdpy11.NOGROWTH
                s = "\t\tGrowth rates reset to {} in {} and {}\n"
                temp = (fwdpy11.NOGROWTH,
                        self._label_deme(e.source),
                        self._label_deme(e.destination))
                self._report.append(s.format(*temp))
            # Even if growth rates are not reset,
            # the onset times and initial sizes are affected
            self.growth_onset_times[e.source] = t
            self.growth_onset_times[e.destination] = t
            self.growth_initial_sizes[e.source] = \
                self.current_deme_sizes[e.source]
            self.growth_initial_sizes[e.destination] = \
                self.current_deme_sizes[e.destination]
            self._report.append("\t\tGrowth onset times changed:\n")
            temp = (t, self._label_deme(e.source))
            self._report.append("\t\t\t{} in deme {}\n".format(*temp))
            temp = (t, self._label_deme(e.destination))
            self._report.append("\t\t\t{} in deme {}\n".format(*temp))
            self._report.append("\t\tGrowth initial sizes changed:\n")
            temp = (self.growth_initial_sizes[e.source],
                    self._label_deme(e.source))
            self._report.append("\t\t\t{} in deme {}\n".format(*temp))
            temp = (self.growth_initial_sizes[e.destination],
                    self._label_deme(e.destination))
            self._report.append("\t\t\t{} in deme {}\n".format(*temp))

    def _apply_SetDemeSize(self, t, event_queues):
        for e in self._current_events(t, event_queues, 'set_deme_sizes'):
            self.current_deme_sizes[e.deme] = e.new_size
            if e.new_size == 0:
                self.went_extinct[e.deme] = 1
            temp = (e.new_size, self._label_deme(e.deme))
            self._report.append("\tDeme size set to {} "
                                "in deme {}\n".format(*temp))
            if e.resets_growth_rate is True:
                self.growth_rates[e.deme] = fwdpy11.NOGROWTH
                temp = (fwdpy11.NOGROWTH,)
                self._report.append("\t\tGrowth rate set to "
                                    "{}\n".format(*temp))

            # Deme size has change.  So, no matter what,
            # there is a new onsite time for growth!
            self.growth_onset_times[e.deme] = t
            self.growth_initial_sizes[e.deme] = self.current_deme_sizes[e.deme]
            temp = self.growth_initial_sizes[e.deme]
            self._report.append("\t\tGrowth initial size "
                                "set to {}\n".format(temp))

    def _apply_SetSelfingRate(self, t, event_queues):
        for e in self._current_events(t, event_queues, 'set_selfing_rates'):
            if self.current_deme_sizes[e.deme] == 0:
                temp = (self._label_deme(e.deme), e.when)
                raise ValueError("Setting selfing rate in "
                                 "extinct deme {} at "
                                 "time {}".format(*temp))
            self.selfing_rates[e.deme] = e.S
            temp = (e.S, self._label_deme(e.deme))
            self._report.append("\tSelfing probability "
                                "set to {} in deme {}\n".format(*temp))

    def _apply_SetExponentialGrowth(self, t, event_queues):
        for e in self._current_events(t, event_queues, 'set_growth_rates'):
            if self.current_deme_sizes[e.deme] == 0:
                temp = (self._label_deme(e.deme), e.when)
                raise ValueError(
                    "attempt to change growth "
                    "rate in extinct deme {} "
                    "at time {}".format(*temp))
            self.growth_rates[e.deme] = e.G
            self.growth_onset_times[e.deme] = t
            self.growth_initial_sizes[e.deme] = self.current_deme_sizes[e.deme]
            temp = (self._label_deme(e.deme), e.G)
            self._report.append("\tGrowth rate "
                                "in deme {} set to {}\n".format(*temp))
            self._report.append("\t\tOnset time: {}\n".format(t))
            N = self.growth_initial_sizes[e.deme]
            self._report.append("\t\tOnset size: {}\n".format(N))

    def _apply_SetMigrationRates(self, t, event_queues):
        for e in self._current_events(t, event_queues, 'set_migration_rates'):
            if e.deme >= 0:
                self.M[e.deme, :] = e.migrates
                temp = (self._label_deme(e.deme), e.migrates)
                self._report.append("\tMigration rates into "
                                    "deme {} set to {}\n".format(*temp))
            else:
                self.M[:] = e.migrates.reshape(self.M.shape)
                self._report.append("\tMigration matrix "
                                    "reset to:\n\t\t{}".format(self.M))

    def _apply_growth_rates(self, t, event_queues):
        next_deme_sizes = np.copy(self.current_deme_sizes)
        for i, j in enumerate(self.growth_rates):
            if j != fwdpy11.NOGROWTH:
                if self.current_deme_sizes[i] == 0:
                    temp = (t, self._label_deme(i))
                    raise ValueError("growth is happening "
                                     "at time {} in extinct "
                                     "deme {}".format(*temp))
                onset = self.growth_onset_times[i]
                G = self.growth_rates[i]
                Nonset = self.growth_initial_sizes[i]
                nextN = np.rint(Nonset * np.power(G, t - onset+1)).astype(int)
                if nextN <= 0:
                    nextN = 0
                    self.growth_rates[i] = fwdpy11.NOGROWTH
                    self.growth_initial_sizes[i] = 0
                    self.growth_onset_times[i] = t
                    self.went_extinct[i] = 1
                next_deme_sizes[i] = nextN

        return next_deme_sizes

    def _validate_migraton_rates(self, t, next_deme_sizes):
        if self.M is None:
            return
        for i, j in enumerate(next_deme_sizes):
            if j == 0:
                if self.M[i, ].sum() != 0:
                    temp = (self._label_deme(i),
                            t, self.M[i, j])
                    raise ValueError("there is migration "
                                     "into empty deme {} "
                                     "at time {}, "
                                     "migrates={}".format(*temp))

    def _check_for_valid_parents(self, t):
        for i, j in enumerate(self.has_metadata):
            if j == 0 and self.current_deme_sizes[i] > 0:
                N = self.current_deme_sizes[i]
                if self.M is None:
                    s = "deme {} has size {} at time {} " + \
                        "but has no valid parents"
                    temp = (self._label_deme(i), N, t)
                    raise ValueError(s.format(*temp))
                # FIXME: if M[i, i] == 0 and M[i, ].sum() > 0,
                # then deme i will have metadata.
                elif self.M[i, i] != 0.0:
                    s = "deme {} at time {} " + \
                        "has no valid parents in that deme " + \
                        "but M[i, i] != 0"
                    temp = (self._label_deme(i), N, t)
                    raise ValueError(s.format(*temp))
                elif self.M[i, ].sum() > 0:
                    self.has_metadata[i] = 1
                    self.went_extinct[i] = 0

    def _generate_report(self, event_queues, simlen):
        """
        Apply events in the same way as the C++
        back-end.
        """
        temp = self._format_deme_sizes(self.current_deme_sizes)
        self._report = [
            "Deme sizes at time {}: {}\n".format(0, temp)
        ]
        t = self._get_next_event_time(event_queues)
        global_extinction = False
        last_t = None
        while t is not None and global_extinction is False:
            self.has_metadata[(self.went_extinct == 1)] = 0
            self.went_extinct[:] = 0
            self.current_deme_sizes[:] = self._apply_growth_rates(
                t-1, event_queues)
            self._report.append("Events at time {}:\n".format(t))
            self._apply_MassMigration(t, event_queues)
            self._apply_SetDemeSize(t, event_queues)
            self._apply_SetExponentialGrowth(t, event_queues)
            self._apply_SetSelfingRate(t, event_queues)
            self._apply_SetMigrationRates(t, event_queues)
            next_deme_sizes = self._apply_growth_rates(t, event_queues)
            self._check_for_valid_parents(t)
            temp = self._format_deme_sizes(next_deme_sizes)
            sizes = "\tDeme sizes after growth: {}\n".format(temp)
            self._report.append(sizes)
            extinct = "\tThe following demes went extinct: {}\n"
            e = np.where(self.went_extinct == 1)[0]
            if len(e) > 0:
                self._report.append(extinct.format(e))
            extinct = "\tThe following demes are extinct: {}\n"
            e = np.where(self.has_metadata == 0)[0]
            if len(e) > 0:
                self._report.append(extinct.format(e))
            if next_deme_sizes.sum() == 0:
                global_extinction = True
                temp = "Global extinction occurs at time {}".format(t)
                warnings.warn(temp)
                self._report.append(temp + '\n')
            self._validate_migraton_rates(t, next_deme_sizes)
            self.current_deme_sizes = next_deme_sizes
            last_t = t
            t = self._get_next_event_time(event_queues)

        if simlen is not None and simlen > last_t:
            deme_sizes = np.copy(self.current_deme_sizes)
            for i, j in enumerate(deme_sizes):
                if j > 0:
                    G = self.growth_rates[i]
                    if G != fwdpy11.NOGROWTH:
                        N0 = self.growth_initial_sizes[i]
                        t0 = self.growth_onset_times[i]
                        N = np.rint(
                            N0*np.power(G, simlen-t0)).astype(int)
                        deme_sizes[i] = N
            temp = "Final deme sizes at time {}: {}"
            deme_sizes = self._format_deme_sizes(deme_sizes)
            self._report.append(temp.format(simlen, deme_sizes))

    def _process_demographic_model(self, events, simlen):
        event_queues = self._make_event_queues(events)
        self._generate_report(event_queues, simlen)

    @property
    def report(self):
        """
        Obtain the details of the demographic
        model as a nicely-formatting string.
        """
        return str('').join(self._report)
